###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This modules builds a widget to visualize workflow execution logs """
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.log.module_exec import ModuleExec
from vistrails.core.log.group_exec import GroupExec
from vistrails.core.log.loop_exec import LoopExec, LoopIteration
from vistrails.core.log.workflow_exec import WorkflowExec
from vistrails.gui.pipeline_view import QPipelineView
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface
from vistrails.gui.collection.workspace import QWorkspaceWindow
from vistrails.core import system, debug
import vistrails.core.db.io


##############################################################################


class QExecutionItem(QtGui.QTreeWidgetItem):
    """
    QExecutionItem represents a workflow or module execution.

    """
    def __init__(self, execution, parent=None, prev=None):
        QtGui.QTreeWidgetItem.__init__(self, parent)
        self.execution = execution
        execution.item = self
        self.modules = []
        self.wf_item = prev or self

        if isinstance(execution, WorkflowExec):
            for item_exec in execution.item_execs:
                QExecutionItem(item_exec, self, self)
            if execution.completed == -2:
                brush = CurrentTheme.SUSPENDED_MODULE_BRUSH
            elif execution.completed == 1:
                brush = CurrentTheme.SUCCESS_MODULE_BRUSH
            else:
                brush = CurrentTheme.ERROR_MODULE_BRUSH

            if execution.db_name:
                self.setText(0, execution.db_name)
            else:
                self.setText(0, 'Version #%s' % execution.parent_version )
        elif isinstance(execution, ModuleExec):
            prev.modules.append(self)
            for loop_exec in execution.loop_execs:
                QExecutionItem(loop_exec, self, prev)
            if execution.completed == 1:
                if execution.error:
                    brush = CurrentTheme.ERROR_MODULE_BRUSH
                    self.wf_item.execution.completed = -1
                elif execution.cached:
                    brush = CurrentTheme.NOT_EXECUTED_MODULE_BRUSH
                else:
                    brush = CurrentTheme.SUCCESS_MODULE_BRUSH
            elif execution.completed == -2:
                brush = CurrentTheme.SUSPENDED_MODULE_BRUSH
            else:
                brush = CurrentTheme.ERROR_MODULE_BRUSH
            self.setText(0, '%s' % execution.module_name)
        elif isinstance(execution, GroupExec):
            prev.modules.append(self)
            for item_exec in execution.item_execs:
                if isinstance(item_exec, LoopExec):
                    QExecutionItem(item_exec, self, prev)
                else:
                    QExecutionItem(item_exec, self, self)
            if execution.completed == 1:
                if execution.error:
                    self.wf_item.execution.completed = -1
                    brush = CurrentTheme.ERROR_MODULE_BRUSH
                elif execution.cached:
                    brush = CurrentTheme.NOT_EXECUTED_MODULE_BRUSH
                else:
                    brush = CurrentTheme.SUCCESS_MODULE_BRUSH
            elif execution.completed == -2:
                brush = CurrentTheme.SUSPENDED_MODULE_BRUSH
            else:
                brush = CurrentTheme.ERROR_MODULE_BRUSH
            self.setText(0, 'Group')
        elif isinstance(execution, LoopExec):
            for iteration in execution.loop_iterations:
                QExecutionItem(iteration, self, prev)
            brush = CurrentTheme.MODULE_BRUSH
            self.setText(0, 'Loop')
        elif isinstance(execution, LoopIteration):
            for item_exec in execution.item_execs:
                QExecutionItem(item_exec, self, prev)
            if execution.completed == 1:
                if execution.error:
                    self.wf_item.execution.completed = -1
                    brush = CurrentTheme.ERROR_MODULE_BRUSH
                else:
                    brush = CurrentTheme.SUCCESS_MODULE_BRUSH
            elif execution.completed == -2:
                brush = CurrentTheme.SUSPENDED_MODULE_BRUSH
            else:
                brush = CurrentTheme.ERROR_MODULE_BRUSH
            self.setText(0, 'Iteration #%s' % execution.iteration)

        self.setText(1, execution.ts_start.strftime(
                    '%H:%M %d/%m').replace(' 0', ' ').replace('/0', '/'))
        self.setData(1, QtCore.Qt.UserRole, str(execution.ts_start))
        #self.setText(2, '%s' % execution.ts_end) end is hidden
        pixmap = QtGui.QPixmap(10,10)
        pixmap.fill(brush.color())
        icon = QtGui.QIcon(pixmap)
        self.setIcon(0, icon)


    def __lt__( self, other ):

        tree = self.treeWidget()
        if ( not tree ):
            column = 0
        else:
            column = tree.sortColumn()

        if column != 1: # only use special sorting for date
            return super(QExecutionItem, self).__lt__(other)

        return self.data(1, QtCore.Qt.UserRole) < other.data(1, QtCore.Qt.UserRole)

class QExecutionListWidget(QtGui.QTreeWidget):
    """
    QExecutionListWidget is a widget containing a list of workflow executions.
    
    """
    def __init__(self, parent=None):
        QtGui.QTreeWidget.__init__(self, parent)
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.setColumnCount(2)
        self.setHeaderLabels(['Pipeline', 'Start']) # end is hidden
        self.header().setDefaultSectionSize(200)
        self.sortByColumn(1, QtCore.Qt.AscendingOrder)
        self.setSortingEnabled(True)

    def set_log(self, log=None):
        self.clear()
        if log is not None:
            for execution in log:
                self.addTopLevelItem(QExecutionItem(execution))

    def add_workflow_exec(self, workflow_exec):
        # mark as recent
        workflow_exec.db_name = workflow_exec.db_name + '*' \
                             if workflow_exec.db_name \
          else '%s*' % self.controller.get_pipeline_name(
                  workflow_exec.parent_version)
        
        self.addTopLevelItem(QExecutionItem(workflow_exec))
       
    
class QLegendBox(QtGui.QFrame):
    """
    QLegendBox is just a rectangular box with a solid color
    
    """
    def __init__(self, brush, size, parent=None, f=QtCore.Qt.WindowFlags()):
        """ QLegendBox(color: QBrush, size: (int,int), parent: QWidget,
                      f: WindowFlags) -> QLegendBox
        Initialize the widget with a color and fixed size
        
        """
        QtGui.QFrame.__init__(self, parent, f)
        self.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Plain)
        self.setAttribute(QtCore.Qt.WA_PaintOnScreen)
        self.setAutoFillBackground(True)
        palette = QtGui.QPalette(self.palette())
        palette.setBrush(QtGui.QPalette.Window, brush)
        self.setPalette(palette)
        self.setFixedSize(*size)
        if system.systemType in ['Darwin']:
            #the mac's nice looking messes up with the colors
            if QtCore.QT_VERSION < 0x40500:
                self.setAttribute(QtCore.Qt.WA_MacMetalStyle, False)
            else:
                self.setAttribute(QtCore.Qt.WA_MacBrushedMetal, False)
        

class QLegendWidget(QtGui.QWidget):
    """
    QLegendWindow contains a list of QLegendBox and its description
    
    """
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.gridLayout = QtGui.QGridLayout(self)
        self.gridLayout.setMargin(10)
        self.gridLayout.setSpacing(10)
        self.setFont(CurrentTheme.VISUAL_DIFF_LEGEND_FONT)
        
        data = [[0, 0, "Successful",      CurrentTheme.SUCCESS_MODULE_BRUSH],
                [0, 1, "Error",             CurrentTheme.ERROR_MODULE_BRUSH],
                [0, 2, "Cached",     CurrentTheme.NOT_EXECUTED_MODULE_BRUSH],
                [1, 0, "Not executed", CurrentTheme.PERSISTENT_MODULE_BRUSH],
                [1, 1, "Suspended",     CurrentTheme.SUSPENDED_MODULE_BRUSH]]

        for x, y, text, brush in data:         
            self.gridLayout.addWidget(
                QLegendBox(brush, CurrentTheme.VISUAL_DIFF_LEGEND_SIZE, self),
                x*2, y*2)
            self.gridLayout.addWidget(QtGui.QLabel(text, self), x*2, y*2+1)

class QLogDetails(QtGui.QWidget, QVistrailsPaletteInterface):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.controller = None
        self.execution = None
        self.parentItem = None
        self.set_title("Log Details")
        self.legend = QLegendWidget()
        self.executionList = QExecutionListWidget()
        self.executionList.setExpandsOnDoubleClick(False)
        self.isDoubling = False
        self.isUpdating = False
        layout = QtGui.QVBoxLayout()

        self.backButton = QtGui.QPushButton('Go back')
        self.backButton.setToolTip("Go back to parent workflow")
        layout.addWidget(self.backButton)
        self.backButton.hide()

        layout.addWidget(self.legend)
        layout.addWidget(self.executionList)
        self.detailsWidget = QtGui.QTextEdit()
        layout.addWidget(self.detailsWidget)
        self.setLayout(layout)
        self.connect(self.executionList, 
                     QtCore.SIGNAL("itemSelectionChanged()"),
                     self.set_execution)
        self.connect(self.backButton,
                     QtCore.SIGNAL('clicked()'),
                     self.goBack)
#        self.connect(self.executionList, QtCore.SIGNAL(
#         "itemClicked(QTreeWidgetItem *, int)"),
#         self.singleClick)
        self.connect(self.executionList, QtCore.SIGNAL(
         "itemDoubleClicked(QTreeWidgetItem *, int)"),
         self.doubleClick)
        self.addButtonsToToolbar()

    def addButtonsToToolbar(self):
        # Add the open version action
        self.openVersionAction = QtGui.QAction(
            QtGui.QIcon.fromTheme('go-next'),
            'Go to this pipeline', None, triggered=self.openVersion)
        self.toolWindow().toolbar.insertAction(self.toolWindow().pinAction,
                                               self.openVersionAction)

    def openVersion(self):

        parent = self.executionList.selectedItems()
        if len(parent) != 1:
            return
        parent = parent[0]
        while not isinstance(parent.execution, WorkflowExec):
            parent = parent.parent()

        version = parent.execution.parent_version
        from vistrails.gui.vistrails_window import _app
        _app.get_current_view().version_selected(version, True)
        self.controller.recompute_terse_graph()
        _app.get_current_view().version_view.select_current_version()
        _app.get_current_view().version_view.scene().setupScene(self.controller)
        _app.qactions['pipeline'].trigger()

    def execution_updated(self):
        for e in self.controller.log.workflow_execs:
            if e not in self.log:
                self.log[e] = e
                wf_id = e.parent_version
                tagMap = self.controller.vistrail.get_tagMap()
                if wf_id in tagMap:
                    e.db_name = tagMap[wf_id]
                self.executionList.add_workflow_exec(e)
                       
    def set_execution(self):
        item = self.executionList.selectedItems()
        if not item:
            return
        item = item[0]
        if self.isDoubling:
            self.isDoubling = False
            return
        if isinstance(item.wf_item, GroupExec):
            self.backButton.show()
        else:
            self.backButton.hide()
        self.notify_app(item.wf_item, item.execution)

    def notify_app(self, wf_item, execution):
        # make sure it is only called once
        if self.isUpdating:
            return
        self.isUpdating = True
        from vistrails.gui.vistrails_window import _app
        _app.notify("execution_changed", wf_item, execution)
        self.isUpdating = False

    def set_controller(self, controller):
        #print '@@@@ QLogDetails calling set_controller'
        if self.controller == controller:
            return

        self.controller = controller
        self.executionList.controller = self.controller
        if self.controller is not None:
            if not hasattr(self.controller, 'loaded_workflow_execs'):
                self.controller.loaded_workflow_execs = {}
                for e in self.controller.read_log().workflow_execs:
                    # set workflow names
                    e.db_name = controller.get_pipeline_name(e.parent_version)
                    self.controller.loaded_workflow_execs[e] = e
            self.log = self.controller.loaded_workflow_execs
        else:
            self.log = None
        self.executionList.set_log(self.log)

    def execution_changed(self, wf_item, execution):
        if not execution:
            return
        self.execution = execution
        self.parentItem = wf_item
        text = ''
        if hasattr(execution, 'item') and \
           not execution.item == self.executionList.currentItem():
            self.executionList.setCurrentItem(execution.item)
        if hasattr(execution, 'item'):
            text += '%s\n' % execution.item.text(0)
        text += 'Start: %s\n' % execution.ts_start
        text += 'End: %s\n' % execution.ts_end
        if hasattr(execution, 'user'):
            text += 'User: %s\n' % execution.user
        if hasattr(execution, 'cached'):
            text += 'Cached: %s\n' % ("Yes" if execution.cached else 'No')
        if hasattr(execution, 'completed'):
            text += 'Completed: %s\n' % {'0':'No', '1':'Yes'}.get(
                                        str(execution.completed), 'No')
        if hasattr(execution, 'error') and execution.error:
            text += 'Error: %s\n' % execution.error
        annotations = execution.db_annotations \
                      if hasattr(execution, 'db_annotations') else []
        if len(annotations):
            text += '\n\nAnnotations:'
            for annotation in annotations:
                text += "\n  %s: %s" % (annotation.key, annotation.value)
        self.detailsWidget.setText(text)
        self.update_selection()
        
    def singleClick(self, item, col):
        if self.isDoubling:
            self.isDoubling = False
            return
        if isinstance(item.wf_item, GroupExec):
            self.backButton.show()
        else:
            self.backButton.hide()
        self.notify_app(item.wf_item, item.execution)

    def doubleClick(self, item, col):
        # only difference here is that we should show contents of GroupExecs 
        self.isDoubling = True
        if isinstance(item.wf_item, GroupExec):
            self.backButton.show()
        else:
            self.backButton.hide()
        if isinstance(item.execution, GroupExec):
            # use itself as the workflow
            self.notify_app(item.execution, item.execution)
        else:
            self.notify_app(item.wf_item, item.execution)

    def goBack(self):
        if not isinstance(self.parentItem.execution, GroupExec):
            self.backButton.hide()
        self.notify_app(self.parentItem.item.wf_item,
                        self.parentItem)

    def update_selection(self):
        if hasattr(self.execution, 'item') and \
          not self.execution.item.isSelected():
            self.execution.item.setSelected(True)

class QLogView(QPipelineView):
    def __init__(self, parent=None):
        QPipelineView.__init__(self, parent)
        self.setReadOnlyMode(True)
        self.set_title("Provenance")
        self.log = None
        self.execution = None
        self.parentItem = None
        self.isUpdating = False
        # Hook shape selecting functions
        self.connect(self.scene(), QtCore.SIGNAL("moduleSelected"),
                     self.moduleSelected)

    def set_default_layout(self):
        self.set_palette_layout({QtCore.Qt.RightDockWidgetArea: QLogDetails})
            
    def set_action_links(self):
        self.action_links = { }
        
    def set_action_defaults(self):
        self.action_defaults.update(
            {'execute' : [('setEnabled', False, False)],
             'publishWeb': [('setEnabled', False, False)],
             'publishPaper': [('setEnabled', False, False)],
            })

    def notify_app(self, wf_item, execution):
        # make sure it is only called once
        if self.isUpdating:
            return
        self.isUpdating = True
        from vistrails.gui.vistrails_window import _app
        _app.notify("execution_changed", wf_item, execution)
        self.isUpdating = False


    def set_controller(self, controller):
        QPipelineView.set_controller(self, controller)
        if not hasattr(self.controller, 'loaded_workflow_execs'):
            self.controller.loaded_workflow_execs = {}
            for e in self.controller.read_log().workflow_execs:
                # set workflow names
                e.db_name = controller.get_pipeline_name(e.parent_version)
                self.controller.loaded_workflow_execs[e] = e
        self.log = self.controller.loaded_workflow_execs

    def set_to_current(self):
        self.controller.set_pipeline_view(self)

    def version_changed(self):
        pass

    def moduleSelected(self, id, selectedItems):
        """ moduleSelected(id: int, selectedItems: [QGraphicsItem]) -> None
        """
        if len(selectedItems)!=1 or id==-1:
            if self.execution != self.parentItem.execution:
                self.notify_app(self.parentItem, self.parentItem.execution)
            return

        item = selectedItems[0]
        if hasattr(item,'execution') and item.execution:
            if self.execution != item.execution:
                item = self.scene().selectedItems()[0]
                self.notify_app(self.parentItem, item.execution)
        elif self.execution != self.parentItem.execution:
            self.notify_app(self.parentItem, self.parentItem.execution)

    def set_exec_by_id(self, exec_id):
        if not self.log:
            return False
        try:
            workflow_execs = [e for e in self.log 
                                if e.id == int(str(exec_id))]
        except ValueError:
            return False
        if len(workflow_execs):
            self.notify_app(workflow_execs[0].item, workflow_execs[0])
            return True
        return False

    def set_exec_by_date(self, exec_date):
        if not self.log:
            return False
        workflow_execs = [e for e in self.log
                          if str(e.ts_start) == str(exec_date)]
        if len(workflow_execs):
            self.notify_app(workflow_execs[0].item, workflow_execs[0])
            return True
        return False

    def get_execution_pipeline(self, execution):
        """ Recursively finds pipeline through layers of groupExecs """
        if isinstance(execution, WorkflowExec):
            version = execution.parent_version
            # change the current version to this as well

            return self.controller.vistrail.getPipeline(version)
        if isinstance(execution, GroupExec):
            parent = execution.item.wf_item.execution
            parent_pipeline = self.get_execution_pipeline(parent)
            return parent_pipeline.db_get_module_by_id(
                                   execution.db_module_id).pipeline

    def execution_changed(self, wf_item, execution):
        self.execution = execution
        if self.parentItem != wf_item:
            self.parentItem = wf_item
            self.pipeline = self.get_execution_pipeline(wf_item.execution)
            self.update_pipeline()
        self.update_selection()

    def update_pipeline(self):
        scene = self.scene()
        scene.clearItems()
        self.pipeline.validate(False)
        
        modules = [(e.execution.module_id, e.execution) for e in self.parentItem.modules
                                                   if hasattr(e.execution, 'module_id')]
        modules.reverse()
        module_execs = {}
        for id, m in modules:
            if id not in module_execs:
                module_execs[id] = []
            module_execs[id].append(m)

        # controller = DummyController(self.pipeline)
        scene.controller = self.controller
        self.moduleItems = {}
        for m_id in self.pipeline.modules:
            module = self.pipeline.modules[m_id]
            brush = CurrentTheme.PERSISTENT_MODULE_BRUSH
            if m_id in module_execs:
                e = module_execs[m_id][-1]
                if e.completed == 1:
                    if e.error:
                        brush = CurrentTheme.ERROR_MODULE_BRUSH
                    elif e.cached:
                        brush = CurrentTheme.NOT_EXECUTED_MODULE_BRUSH
                    else:
                        brush = CurrentTheme.SUCCESS_MODULE_BRUSH
                elif e.completed == -2:
                    brush = CurrentTheme.SUSPENDED_MODULE_BRUSH
                else:
                    brush = CurrentTheme.ERROR_MODULE_BRUSH
            module.is_valid = True
            item = scene.addModule(module, brush)
            item.controller = self.controller
            self.moduleItems[m_id] = item
            if m_id in module_execs:
                for e in module_execs[m_id]:
                    item.execution = e
                    e.module = item
            else:
                item.execution = None
        connectionItems = []
        for c in self.pipeline.connections.values():
            connectionItems.append(scene.addConnection(c))

        # Color Code connections
        for c in connectionItems:
            pen = QtGui.QPen(CurrentTheme.CONNECTION_PEN)
            pen.setBrush(QtGui.QBrush(QtGui.QColor(0, 0, 0, 128+64)))
            c.connectionPen = pen

        scene.updateSceneBoundingRect()
        scene.fitToView(self, True)

    def update_selection(self):
        # avoid module update signal
        self.isUpdating = True
        module = None
        if (isinstance(self.execution, ModuleExec) or \
            isinstance(self.execution, GroupExec)) and \
            hasattr(self.execution, 'module') and \
          not self.execution.module.isSelected():
            self.execution.module.setSelected(True)
            module = self.execution.module
        for m in self.scene().selectedItems():
            if m.isSelected() and m != module:
                m.setSelected(False)
        self.isUpdating = False



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

# Finds all .py files recursively in current directory (.)
# and updates to modified BSD license in the file header. 
import re
import os
py_bsd_3_license = """###############################################################################
##
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the University of Utah nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
xml_bsd_3_license = "<!--"+ py_bsd_3_license + "-->"
php_bsd_3_license = """////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2006-2011, University of Utah.
// All rights reserved.
// Contact: contact@vistrails.org
//
// This file is part of VisTrails.
//
// "Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  - Neither the name of the University of Utah nor the names of its
//    contributors may be used to endorse or promote products derived from
//    this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
//
////////////////////////////////////////////////////////////////////////////"""
sql_bsd_3_license = """--#############################################################################
--
-- Copyright (C) 2006-2011, University of Utah.
-- All rights reserved.
-- Contact: contact@vistrails.org
--
-- This file is part of VisTrails.
--
-- "Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
--  - Redistributions of source code must retain the above copyright notice,
--    this list of conditions and the following disclaimer.
--  - Redistributions in binary form must reproduce the above copyright
--    notice, this list of conditions and the following disclaimer in the
--    documentation and/or other materials provided with the distribution.
--  - Neither the name of the University of Utah nor the names of its
--    contributors may be used to endorse or promote products derived from
--    this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
-- CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-- EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-- PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
-- OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
-- WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
-- OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
-- ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--
--#############################################################################"""
py_old_license = """############################################################################
##
## Copyright (C) 2006-2010 University of Utah. All rights reserved.
##
## This file is part of VisTrails.
##
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE.GPL included in the packaging of
## this file.  Please review the following to ensure GNU General Public
## Licensing requirements will be met:
## http://www.opensource.org/licenses/gpl-license.php
##
## If you are unsure which license is appropriate for your use (for
## instance, you are interested in developing a commercial derivative
## of VisTrails), please contact us at contact@vistrails.org.
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
############################################################################
"""
sql_old_license = """--#########################################################################
--
-- Copyright (C) 2006-2010 University of Utah. All rights reserved.
--
-- This file is part of VisTrails.
--
-- This file may be used under the terms of the GNU General Public
-- License version 2.0 as published by the Free Software Foundation
-- and appearing in the file LICENSE.GPL included in the packaging of
-- this file.  Please review the following to ensure GNU General Public
-- Licensing requirements will be met:
-- http://www.opensource.org/licenses/gpl-license.php
--
-- If you are unsure which license is appropriate for your use (for
-- instance, you are interested in developing a commercial derivative
-- of VisTrails), please contact us at contact@vistrails.org.
--
-- This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
-- WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
--##########################################################################"""
xml_old_license = """<!--########################################################################
##
## Copyright (C) 2006-2010 University of Utah. All rights reserved.
##
## This file is part of VisTrails.
##
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE.GPL included in the packaging of
## this file.  Please review the following to ensure GNU General Public
## Licensing requirements will be met:
## http://www.opensource.org/licenses/gpl-license.php
##
## If you are unsure which license is appropriate for your use (for
## instance, you are interested in developing a commercial derivative
## of VisTrails), please contact us at contact@vistrails.org.
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
#########################################################################-->"""

php_old_license = """////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2006-2010 University of Utah. All rights reserved.
//
// This file is part of VisTrails.
//
// This file may be used under the terms of the GNU General Public
// License version 2.0 as published by the Free Software Foundation
// and appearing in the file LICENSE.GPL included in the packaging of
// this file.  Please review the following to ensure GNU General Public
// Licensing requirements will be met:
// http://www.opensource.org/licenses/gpl-license.php
//
// If you are unsure which license is appropriate for your use (for
// instance, you are interested in developing a commercial derivative
// of VisTrails), please contact us at contact@vistrails.org.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
////////////////////////////////////////////////////////////////////////////"""
def process_files(files, old_license, new_license, type):
    count = 0
    not_found = []
    gpl_found = []
    for fname in files:
        fin = open(fname)
        all_lines = fin.read()
        fin.close()
        pos = all_lines.find(old_license)
        pos_gen = all_lines.find("GNU General Public")
        if pos > -1:
            print "Updating: %s"%fname
            newlines = all_lines.replace(old_license, new_license)
            fout = file(fname, 'w')
            fout.write(newlines)
            fout.close()
            count += 1
        else:
            if pos_gen > -1:
                gpl_found.append(fname)
            else:
                if all_lines.find(new_license) == -1:
                    print "Appending license to %s"%fname
                    newlines = new_license+"\n"+all_lines
                    fout = file(fname, 'w')
                    fout.write(newlines)
                    fout.close()
                    count +=1
    print count, " %s files updated"%type
    print " found reference to gpl in the following files:"
    for fname in gpl_found:
        print "  ", fname

py_files = []
xml_files = []
php_files = []
sql_files = []
others = []
for (path, dnames, fnames) in os.walk('.'):
    for fn in fnames:
        if fn.endswith(".py"):
            py_files.append(os.path.join(path, fn))
        elif fn.endswith(".xml") or fn.endswith(".xsd"):
            xml_files.append(os.path.join(path,fn))
        elif fn.endswith(".php"):
            php_files.append(os.path.join(path,fn))
        elif fn.endswith(".sql"):
            sql_files.append(os.path.join(path,fn))
        else:
            others.append(os.path.join(path,fn))

print len(py_files), " python files found"
process_files(py_files, py_old_license, py_bsd_3_license, 'python')

print len(sql_files), " sql files found"
process_files(sql_files, sql_old_license, sql_bsd_3_license, 'sql')

print len(xml_files), " xml files found"
process_files(xml_files, xml_old_license, xml_bsd_3_license, 'xml')

print len(php_files), " php files found"
process_files(php_files, php_old_license, php_bsd_3_license, 'php')

#print "Ignored files: "
#for f in others:
#    print f



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
import re
import os.path

import vtk

from vistrails.core.configuration import ConfigField
from vistrails.core.modules.basic_modules import PathObject, \
                                                       identifier as basic_pkg
from vistrails.core.modules.config import ModuleSettings
from vistrails.core.modules.vistrails_module import ModuleError
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.output_modules import OutputModule, ImageFileMode, \
    ImageFileModeConfig, IPythonMode, IPythonModeConfig
from vistrails.core.system import get_vistrails_default_pkg_prefix, systemType, current_dot_vistrails
from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler,\
                                       UpgradeModuleRemap, UpgradePackageRemap
from vistrails.core.vistrail.connection import Connection
from .pythonclass import BaseClassModule, gen_class_module

from .tf_widget import _modules as tf_modules
from .inspectors import _modules as inspector_modules
from .offscreen import _modules as offscreen_modules

from identifiers import identifier, version as package_version

from .vtk_wrapper import vtk_classes
from . import hasher


_modules = tf_modules + inspector_modules + offscreen_modules

registry = get_module_registry()
if registry.has_module('org.vistrails.vistrails.spreadsheet', 'SpreadsheetCell'):
    # load these only if spreadsheet is enabled
    from .vtkcell import _modules as cell_modules
    from .vtkhandler import _modules as handler_modules
    _modules += cell_modules + handler_modules


################# OUTPUT MODULES #############################################

def render_to_image(output_filename, vtk_format, renderer, w, h):
    window = vtk.vtkRenderWindow()
    window.OffScreenRenderingOn()
    window.SetSize(w, h)

    # FIXME think this may be fixed in VTK6 so we don't have this
    # dependency...
    widget = None
    if systemType=='Darwin':
        from PyQt4 import QtCore, QtGui
        widget = QtGui.QWidget(None, QtCore.Qt.FramelessWindowHint)
        widget.resize(w, h)
        widget.show()
        window.SetWindowInfo(str(int(widget.winId())))

    window.AddRenderer(renderer)
    window.Render()
    win2image = vtk.vtkWindowToImageFilter()
    win2image.SetInput(window)
    win2image.Update()
    writer = vtk_format()
    writer.SetInput(win2image.GetOutput())
    writer.SetFileName(output_filename)
    writer.Write()
    window.Finalize()
    if widget!=None:
        widget.close()

class vtkRendererToFile(ImageFileMode):
    config_cls = ImageFileModeConfig
    formats = ['png', 'jpg', 'tif', 'pnm']

    @classmethod
    def can_compute(cls):
        return True

    def compute_output(self, output_module, configuration):
        format_map = {'png': vtk.vtkPNGWriter,
                      'jpg': vtk.vtkJPEGWriter,
                      'tif': vtk.vtkTIFFWriter,
                      'pnm': vtk.vtkPNMWriter}
        r = output_module.get_input("value")[0].vtkInstance
        w = configuration["width"]
        h = configuration["height"]
        img_format = self.get_format(configuration)
        if img_format not in format_map:
            raise ModuleError(output_module,
                              'Cannot output in format "%s"' % img_format)
        fname = self.get_filename(configuration, suffix='.%s' % img_format)

        render_to_image(fname, format_map[img_format], r, w, h)

class vtkRendererToIPythonModeConfig(IPythonModeConfig):
    _fields = [ConfigField('width', 640, int),
               ConfigField('height', 480, int)]

class vtkRendererToIPythonMode(IPythonMode):
    config_cls = vtkRendererToIPythonModeConfig

    def compute_output(self, output_module, configuration=None):
        from IPython.core.display import display, Image

        r = output_module.get_input('value')[0].vtkInstance
        width = configuration['width']
        height = configuration['height']

        window = vtk.vtkRenderWindow()
        window.OffScreenRenderingOn()
        window.SetSize(width, height)

        fname = output_module.interpreter.filePool.create_file(
                prefix='ipython_', suffix='.png').name

        render_to_image(fname, vtk.vtkPNGWriter, r, width, height)
        display(Image(filename=fname, width=width, height=height))

class vtkRendererOutput(OutputModule):
    _settings = ModuleSettings(configure_widget="vistrails.gui.modules."
                       "output_configuration:OutputModuleConfigurationWidget")
    _input_ports = [('value', 'vtkRenderer', {'depth':1}),
                    ('interactorStyle', 'vtkInteractorStyle'),
                    ('picker', 'vtkAbstractPicker')]
    _output_modes = [vtkRendererToFile, vtkRendererToIPythonMode]
    if registry.has_module('org.vistrails.vistrails.spreadsheet',
                           'SpreadsheetCell'):
        from .vtkcell import vtkRendererToSpreadsheet
        _output_modes.append(vtkRendererToSpreadsheet)

_modules.append(vtkRendererOutput)


################# ADD VTK CLASSES ############################################


# keep track of created modules for use as subclasses
klasses = {}

def initialize():
    # First check if spec for this VTK version exists
    v = vtk.vtkVersion()
    vtk_version = [v.GetVTKMajorVersion(),
                   v.GetVTKMinorVersion(),
                   v.GetVTKBuildVersion()]

    # vtk-VTKVERSION-spec-PKGVERSION.xml
    spec_name = os.path.join(current_dot_vistrails(),
                             'vtk-%s-spec-%s.xml' %
                             ('_'.join([str(v) for v in vtk_version]),
                              package_version.replace('.', '_')))
    # TODO: how to patch with diff/merge
    if not os.path.exists(spec_name):
        from .vtk_wrapper.parse import parse
        parse(spec_name)
    vtk_classes.initialize(spec_name)
    _modules.insert(0, BaseClassModule)
    _modules.extend([gen_class_module(spec, vtk_classes, klasses, signature=hasher.vtk_hasher)
                     for spec in vtk_classes.specs.module_specs])

################# UPGRADES ###################################################

_remap = None
_controller = None
_pipeline = None

def _get_controller():
    global _controller
    return _controller

def _get_pipeline():
    global _pipeline
    return _pipeline

module_name_remap = {'vtkPLOT3DReader': 'vtkMultiBlockPLOT3DReader'}

def base_name(name):
    """Returns name without overload index.
    """
    i = name.find('_')
    if i != -1:
        return name[:i]
    return name

def build_remap(module_name=None):
    global _remap, _controller

    reg = get_module_registry()
    uscore_num = re.compile(r"(.+)_(\d+)$")

    def create_function(module, *argv, **kwargs):
        controller = _get_controller()
        # create function using the current module version and identifier
        # FIXME: This should really be handled by the upgrade code somehow
        new_desc = reg.get_descriptor_by_name(module.package,
                                              module.name,
                                              module.namespace)
        old_identifier = module.package
        module.package = identifier
        old_package_version = module.version
        module.version = new_desc.package_version
        new_function = controller.create_function(module, *argv, **kwargs)
        module.package = old_identifier
        module.version = old_package_version
        return new_function

    def get_port_specs(descriptor, port_type):
        ports = {}
        for desc in reversed(reg.get_module_hierarchy(descriptor)):
            ports.update(reg.module_ports(port_type, desc))
        return ports

    def get_input_port_spec(module, port_name):
        # Get current desc
        # FIXME: This should really be handled by the upgrade code somehow
        new_desc = reg.get_descriptor_by_name(module.package,
                                              module.name,
                                              module.namespace)
        port_specs = get_port_specs(new_desc, 'input')
        return port_name in port_specs and port_specs[port_name]

    def get_output_port_spec(module, port_name):
        # Get current desc
        new_desc = reg.get_descriptor_by_name(module.package,
                                              module.name,
                                              module.namespace)
        port_specs = get_port_specs(new_desc, 'output')
        return port_name in port_specs and port_specs[port_name]

    def build_function(old_function, new_function_name, new_module):
        controller = _get_controller()
        if len(old_function.parameters) > 0:
            new_param_vals, aliases = \
                zip(*[(p.strValue, p.alias)
                      for p in old_function.parameters])
        else:
            new_param_vals = []
            aliases = []
        new_function = create_function(new_module,
                                       new_function_name,
                                       new_param_vals,
                                       aliases)
        return new_function

    def build_function_remap_method(desc, port_prefix, port_num):
        f_map = {"vtkCellArray": {"InsertNextCell": 3}}

        def remap(old_function, new_module):
            for i in xrange(1, port_num):
                port_name = "%s_%d" % (port_prefix, i)
                port_spec = get_input_port_spec(new_module, port_name)
                old_sigstring = \
                    reg.expand_port_spec_string(old_function.sigstring,
                                                basic_pkg)
                if port_spec.sigstring == old_sigstring:
                    new_function = build_function(old_function, port_name,
                                                  new_module)
                    new_module.add_function(new_function)
                    return []
            port_idx = 1
            if desc.name in f_map:
                if port_prefix in f_map[desc.name]:
                    port_idx =  f_map[desc.name][port_prefix]
            port_name = "%s_%d" % (port_prefix, port_idx)
            new_function = build_function(old_function, port_name, new_module)
            new_module.add_function(new_function)
            return []

        return remap

    def build_remap_method(desc, port_prefix, port_num, port_type):
        # for connection, need to differentiate between src and dst
        if port_type == 'input':
            conn_lookup = Connection._get_destination
            get_port_spec = get_input_port_spec
            idx = 1
        else:
            conn_lookup = Connection._get_source
            get_port_spec = get_output_port_spec
            idx = 0

        def remap(old_conn, new_module):
            create_new_connection = UpgradeWorkflowHandler.create_new_connection
            port = conn_lookup(old_conn)
            pipeline = _get_pipeline()
            modules = [pipeline.modules[old_conn.source.moduleId],
                       pipeline.modules[old_conn.destination.moduleId]]
            modules[idx] = new_module
            ports = [old_conn.source, old_conn.destination]
            for i in xrange(1, port_num):
                port_name = "%s_%d" % (port_prefix, i)
                port_spec = get_port_spec(modules[idx], port_name)
                if port_spec.sigstring == port.signature:
                    ports[idx] = port_name
                    new_conn = create_new_connection(_get_controller(),
                                                     modules[0],
                                                     ports[0],
                                                     modules[1],
                                                     ports[1])
                    return [('add', new_conn)]

            # if get here, just try to use _1 version?
            ports[idx] = "%s_%d" % (port_prefix, 1)
            new_conn = create_new_connection(_get_controller(),
                                             modules[0],
                                             ports[0],
                                             modules[1],
                                             ports[1])
            return [('add', new_conn)]
        return remap

    def process_ports(desc, remap, port_type):
        if port_type == 'input':
            remap_dict_key = 'dst_port_remap'
        else:
            remap_dict_key = 'src_port_remap'
        ports = get_port_specs(desc, port_type)
        port_nums = {}
        for port_name, port_spec in ports.iteritems():
            # FIXME just start at 1 and go until don't find port (no
            # need to track max)?
            search_res = uscore_num.search(port_name)
            if search_res:
                port_prefix = search_res.group(1)
                port_num = int(search_res.group(2))
                if port_prefix not in port_nums:
                    port_nums[port_prefix] = port_num
                elif port_num > port_nums[port_prefix]:
                    port_nums[port_prefix] = port_num
        for port_prefix, port_num in port_nums.iteritems():
            m = build_remap_method(desc, port_prefix, port_num, port_type)
            remap.add_remap(remap_dict_key, port_prefix, m)
            if port_type == 'input':
                m = build_function_remap_method(desc, port_prefix, port_num)
                remap.add_remap('function_remap', port_prefix, m)
        if port_type == 'output' and desc.name in klasses:
            remap.add_remap('src_port_remap', 'self', 'Instance')

    def change_func(name, value):
        def remap(old_func, new_module):
            controller = _get_controller()
            new_function = create_function(new_module, name, [value])
            return [('add', new_function, 'module', new_module.id)]
        return remap

    def change_SetXint(spec):
        # Fix old SetX methods that takes an int representing the enum
        def remap(old_func, new_module):
            controller = _get_controller()
            value = int(old_func.params[0].strValue)
            value = spec.values[0][value]
            new_function = create_function(new_module, spec.name, [value])
            return [('add', new_function, 'module', new_module.id)]
        return remap

    def color_func(name):
        def remap(old_func, new_module):
            controller = _get_controller()
            value = ','.join([p.strValue for p in old_func.params])
            new_function = create_function(new_module, name, [value])
            return [('add', new_function, 'module', new_module.id)]
        return remap

    def file_func(name):
        def remap(old_func, new_module):
            controller = _get_controller()
            value = PathObject(old_func.params[0].strValue)
            new_function = create_function(new_module, name, [value])
            return [('add', new_function, 'module', new_module.id)]
        return remap

    def to_file_func(name):
        # Add Path module as name->File converter
        def remap(old_conn, new_module):
            controller = _get_controller()
            create_new_connection = UpgradeWorkflowHandler.create_new_connection
            pipeline = _get_pipeline()
            module = pipeline.modules[old_conn.source.moduleId]
            x = (module.location.x + new_module.location.x)/2
            y = (module.location.y + new_module.location.y)/2
            path_module = controller.create_module(basic_pkg, 'Path',
                                                   '', x, y)
            conn1 = create_new_connection(controller,
                                          module,
                                          old_conn.source,
                                          path_module,
                                          'name')
            conn2 = create_new_connection(controller,
                                          path_module,
                                          'value',
                                          new_module,
                                          name)
            return [('add', path_module),
                    ('add', conn1),
                    ('add', conn2)]
        return remap

    def wrap_block_func():
        def remap(old_conn, new_module):
            controller = _get_controller()
            create_new_connection = UpgradeWorkflowHandler.create_new_connection
            pipeline = _get_pipeline()
            module1 = pipeline.modules[old_conn.destination.moduleId]
            dest_port = old_conn.destination
            candidates = ['AddInputData_1', 'AddInputData',
                          'SetInputData_1', 'SetInputData',
                          'AddInput', 'SetInput']
            if 'Connection' in old_conn.destination.name:
                _desc = reg.get_descriptor_by_name(identifier,
                                                   module1.name)
                ports = get_port_specs(_desc, 'input')
                for c in candidates:
                    if c in ports:
                        dest_port = c
                        break
            conn = create_new_connection(controller,
                                         new_module,
                                         'StructuredGrid',
                                         module1,
                                         dest_port)
            return [('add', conn)]
        return remap

    def fix_vtkcell_func():
        # Move VTKCell.self -> X.VTKCell to
        # vtkRenderer.Instance -> X.vtkRenderer
        def remap(old_conn, new_module):
            controller = _get_controller()
            create_new_connection = UpgradeWorkflowHandler.create_new_connection
            pipeline = _get_pipeline()
            # find vtkRenderer
            vtkRenderer = None
            for conn in pipeline.connections.itervalues():
                src_module_id = conn.source.moduleId
                dst_module_id = conn.destination.moduleId
                if dst_module_id == old_conn.source.moduleId and \
                   pipeline.modules[src_module_id].name == 'vtkRenderer':
                    vtkRenderer = pipeline.modules[src_module_id]
            if vtkRenderer:
                conn = create_new_connection(controller,
                                             vtkRenderer,
                                             'Instance',
                                             new_module,
                                             'vtkRenderer')
                return [('add', conn)]
            return []
        return remap

    def process_module(desc):
        # 0.9.3 upgrades
        if not desc.name in klasses:
            return
        remap = UpgradeModuleRemap(None, '0.9.3', '0.9.3',
                                   module_name=desc.name)
        process_ports(desc, remap, 'input')
        process_ports(desc, remap, 'output')
        _remap.add_module_remap(remap)
        for old, new in module_name_remap.iteritems():
            if desc.name == new:
                # Remap using old name
                remap.new_module = old
                _remap.add_module_remap(remap, old)
        # 0.9.5 upgrades
        remap = UpgradeModuleRemap('0.9.3', '0.9.5', '0.9.5',
                                   module_name=desc.name)
        remap.add_remap('src_port_remap', 'self', 'Instance')
        _remap.add_module_remap(remap)
        for old, new in module_name_remap.iteritems():
            if desc.name == new:
                # Remap using old name
                remap.new_module = old
                _remap.add_module_remap(remap, old)
        # 1.0.0 upgrades
        input_mappings = {}
        function_mappings = {}
        input_specs = [desc.module._get_input_spec(s)
                     for s in get_port_specs(desc, 'input')]
        input_names = [s.name for s in input_specs]
        for spec in input_specs:
            if spec is None:
                continue
            elif spec.name == 'TextScaleMode':
                function_mappings['ScaledTextOn'] = \
                                           change_func('TextScaleMode', 'Prop')
            elif spec.method_type == 'OnOff':
                # Convert On/Off to single port
                input_mappings[spec.name + 'On'] = spec.name
                input_mappings[spec.name + 'Off'] = spec.name
                function_mappings[spec.name + 'On'] = \
                                              change_func(spec.name, True)
                function_mappings[spec.name + 'Off'] = \
                                             change_func(spec.name, False)
            elif spec.method_type == 'nullary':
                # Add True to execute empty functions
                function_mappings[spec.name] = change_func(spec.name, True)
            elif spec.method_type == 'SetXToY':
                # Add one mapping for each default
                for enum in spec.values[0]:
                    input_mappings[spec.method_name + enum] = spec.name
                    # Add enum value to function
                    function_mappings[spec.method_name + enum] = \
                                                  change_func(spec.name, enum)
                # Convert SetX(int) methods
                old_name = spec.method_name[:-2]
                function_mappings[spec.method_name[:-2]] = change_SetXint(spec)
            elif spec.port_type == 'basic:Color':
                # Remove 'Widget' suffix on Color
                input_mappings[spec.method_name + 'Widget'] = spec.name
                # Remove 'Set prefix'
                input_mappings[spec.method_name] = spec.name
                # Change old type (float, float, float) -> (,)*3
                function_mappings[spec.method_name] = color_func(spec.name)
            elif spec.port_type == 'basic:File':
                input_mappings[spec.method_name] = to_file_func(spec.name)  # Set*FileName -> (->File->*File)
                input_mappings['Set' + spec.name] = spec.name # Set*File -> *File
                function_mappings[spec.method_name] = file_func(spec.name)
            elif base_name(spec.name) == 'AddDataSetInput':
                # SetInput* does not exist in VTK 6
                if spec.name[15:] == '_1':
                    # Upgrade from version without overload
                    input_mappings['AddInput'] = spec.name
                input_mappings['AddInput' + spec.name[15:]] = spec.name
            elif base_name(spec.name) == 'InputData':
                # SetInput* does not exist in VTK 6
                if spec.name[9:] == '_1':
                    # Upgrade from version without overload
                    input_mappings['SetInput'] = spec.name
                input_mappings['SetInput' + spec.name[9:]] = spec.name
            elif base_name(spec.name) == 'AddInputData':
                # AddInput* does not exist in VTK 6
                if spec.name[12:] == '_1':
                    # Upgrade from version without overload
                    input_mappings['AddInput'] = spec.name
                input_mappings['AddInput' + spec.name[12:]] = spec.name
            elif base_name(spec.name) ==  'SourceData':
                # SetSource* does not exist in VTK 6
                if spec.name[10:] == '_1':
                    # Upgrade from version without overload
                    input_mappings['SetSource'] = spec.name
                input_mappings['SetSource' + spec.name[10:]] = spec.name
            elif spec.method_name == 'Set' + base_name(spec.name):
                if spec.name[-2:] == '_1':
                    # Upgrade from versions without overload
                    input_mappings[spec.name[:-2]] = spec.name
                    input_mappings['Set' + spec.name[:-2]] = spec.name
                # Remove 'Set' prefixes
                input_mappings['Set' + spec.name] = spec.name
            elif spec.name == 'AddInput_1':
                # FIXME what causes this?
                # New version does not have AddInput
                input_mappings['AddInput'] = 'AddInput_1'
            elif spec.name == 'vtkRenderer':
                # Classes having SetRendererWindow also used to have VTKCell
                input_mappings['SetVTKCell'] = fix_vtkcell_func()
        output_mappings = {}
        for spec_name in get_port_specs(desc, 'output'):
            spec = desc.module._get_output_spec(spec_name)
            if spec is None:
                continue
            if spec.method_name == 'Get' + spec.name:
                # Remove 'Get' prefixes
                output_mappings[spec.method_name] = spec.name
        if desc.name == 'vtkMultiBlockPLOT3DReader':
            # Move GetOutput to custom FirstBlock
            output_mappings['GetOutput'] = wrap_block_func()  # what!?
            # Move GetOutputPort0 to custom FirstBlock
            # and change destination port to AddInputData_1 or similar
            output_mappings['GetOutputPort0'] = wrap_block_func()

        remap = UpgradeModuleRemap('0.9.5', '1.0.0', '1.0.0',
                                   module_name=desc.name)
        for k, v in input_mappings.iteritems():
            remap.add_remap('dst_port_remap', k, v)
        for k, v in output_mappings.iteritems():
            remap.add_remap('src_port_remap', k, v)
        for k, v in function_mappings.iteritems():
            remap.add_remap('function_remap', k, v)
        _remap.add_module_remap(remap)
        for old, new in module_name_remap.iteritems():
            if desc.name == new:
                # Remap to new name
                remap.new_module = new
                _remap.add_module_remap(remap, old)

    pkg = reg.get_package_by_name(identifier)
    if module_name is not None:
        desc = reg.get_descriptor_by_name(identifier, module_name)
        process_module(desc)
    else:
        # FIXME do this by descriptor first, then build the hierarchies for each
        # module after that...
        for desc in pkg.descriptor_list:
            process_module(desc)

def handle_module_upgrade_request(controller, module_id, pipeline):
    global _remap, _controller, _pipeline

    if _remap is None:
        _remap = UpgradePackageRemap()
        remap = UpgradeModuleRemap(None, '1.0.0', '1.0.0',
                                   module_name='vtkInteractionHandler')
        remap.add_remap('src_port_remap', 'self', 'Instance')
        _remap.add_module_remap(remap)
        remap = UpgradeModuleRemap(None, '1.0.0', '1.0.0',
                                   module_name='VTKCell')
        remap.add_remap('src_port_remap', 'self', 'Instance')
        _remap.add_module_remap(remap)
        remap = UpgradeModuleRemap(None, '1.0.0', '1.0.0',
                                   module_name='VTKViewCell',
                                   new_module='VTKCell')
        _remap.add_module_remap(remap)

    _controller = controller
    _pipeline = pipeline
    module_name = pipeline.modules[module_id].name
    module_name = module_name_remap.get(module_name, module_name)
    if not _remap.has_module_remaps(module_name):
        build_remap(module_name)

    try:
        from vistrails.packages.spreadsheet.init import upgrade_cell_to_output
    except ImportError:
        # Manually upgrade to 1.0.1
        if _remap.get_module_remaps(module_name):
            module_remap = copy.copy(_remap)
            module_remap.add_module_remap(
                    UpgradeModuleRemap('1.0.0', '1.0.1', '1.0.1',
                                       module_name=module_name))
        else:
            module_remap = _remap
    else:
        module_remap = upgrade_cell_to_output(
                _remap, module_id, pipeline,
                'VTKCell', 'vtkRendererOutput',
                '1.0.1', 'AddRenderer',
                start_version='1.0.0')
        if _remap.get_module_remaps(module_name):
            remap = module_remap.get_module_upgrade(module_name, '1.0.0')
            if remap is None:
                # Manually upgrade to 1.0.1
                module_remap.add_module_remap(
                        UpgradeModuleRemap('1.0.0', '1.0.1', '1.0.1',
                                           module_name=module_name))

    return UpgradeWorkflowHandler.remap_module(controller, module_id, pipeline,
                                               module_remap)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" Helper functions for parsing and evaluating expressions """

from __future__ import division

################################################################################

def evaluate_expressions(expressions):
    """ evaluate_expressions(expressions: str) -> int/float/string        
    Evaluate the contents of the line edit inside each '$' pair
    
    """

    # FIXME: eval should pretty much never be used
    (base, exps) = parse_expression(str(expressions))
    for e in exps:
        try:
            base = base[:e[0]] + unicode(eval(e[1], None, None)) + base[e[0]:]
        except Exception:
            base = base[:e[0]] + '$' + e[1] + '$' + base[e[0]:]
    return base

def parse_expression(expression):
    """ parse_expression(expression: str) -> output (see below)        
    Parse the mixed expression string into expressions and string
    constants

    Keyword arguments:        
    output - (simplified string, [(pos:exp),(pos:exp),...]
        simplified string: the string without any '$exp$'. All
        $$ will be replace by a single $.
    (pos:exp) - the expression to be computed and where it should be
        inserted back to the simplified string, starting from
        the end of the string.

    """
    import re
    output = expression
    result = []
    p = re.compile(r'\$[^$]+\$')
    e = p.finditer(output)
    if e:
        offset = 0
        for s in e:
            exp = s.group()
            result.append((s.span()[0]-offset, exp[1:len(exp)-1]))
            offset += s.span()[1]-s.span()[0]
        result.reverse()
        output = p.sub('', output)
        output.replace('$$','$')
    return (output, result)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.db.versions.v0_7_0.domain import DBVistrail, DBAction, DBTag, DBModule, \
    DBConnection, DBPortSpec, DBFunction, DBParameter, DBLocation, DBAdd, \
    DBChange, DBDelete, DBAnnotation, DBPort

def translateVistrail(_vistrail):
    vistrail = DBVistrail()

    for _action in _vistrail.db_actions.itervalues():
        ops = []
        for op in _action.db_operations:
            if op.vtType == 'add':
                data = convert_data(op.db_data, op.db_parentObjType,
                                    op.db_parentObjId)
                ops.append(DBAdd(id=op.db_id,
                                 what=op.db_what, 
                                 objectId=op.db_objectId, 
                                 parentObjId=op.db_parentObjId, 
                                 parentObjType=op.db_parentObjType, 
                                 data=data))
            elif op.vtType == 'change':
                data = convert_data(op.db_data, op.db_parentObjType,
                                    op.db_parentObjId)
                ops.append(DBChange(id=op.db_id,
                                    what=op.db_what, 
                                    oldObjId=op.db_oldObjId, 
                                    newObjId=op.db_newObjId,
                                    parentObjId=op.db_parentObjId, 
                                    parentObjType=op.db_parentObjType, 
                                    data=data))
            elif op.vtType == 'delete':
                ops.append(DBDelete(id=op.db_id,
                                    what=op.db_what, 
                                    objectId=op.db_objectId, 
                                    parentObjId=op.db_parentObjId, 
                                    parentObjType=op.db_parentObjType))
        action = DBAction(id=_action.db_id,
                          prevId=_action.db_prevId, 
                          date=_action.db_date, 
                          user=_action.db_user, 
                          operations=ops,
                          annotations=_action.db_annotations.values())
        vistrail.db_add_action(action)

    for _tag in _vistrail.db_tags.itervalues():
        tag = DBTag(id=_tag.db_id,
                    name=_tag.db_name)
        vistrail.db_add_tag(tag)

    vistrail.db_version = '0.7.0'
    return vistrail

module_map = {}
translate_ports = {'SetInputConnection': {None: 'SetInputConnection0'},
                   'GetOutputPort': {None: 'GetOutputPort0'},
                   'SetRenderWindow': {None: 'SetVTKCell'},
                   'SetInteractorStyle': {None: 'InteractorStyle'},
                   'ResetCamera': {None: 'ResetCamera'},
                   'AddPoint': {None: 'AddPoint_1'},
                   'AddRGBPoint': {None: 'AddRGBPoint_1'},
                   'AddHSVPoint': {None: 'AddHSVPoint_1'},
                   'AddInput': {None: 'AddInput',
                                'vtkXYPlotActor': 'AddInput_2'},
                   'SetInput': {None: 'SetInput',
                                'vtkPolyDataNormals': 'SetInput_1',
                                'vtkGlyph3D': 'SetInput_1',
                                'vtkDelaunay2D': 'SetInput_1',
                                'vtkDelaunay3D': 'SetInput_1',
                                'vtkWarpVector': 'SetInput_1',
                                'vtkContourFilter': 'SetInput_1',
                                'vtkTubeFilter': 'SetInput_1',
                                'vtkThresholdPoints': 'SetInput_1',
                                'vtkProbeFilter': 'SetInput_1',
                                'vtkTriangleFilter': 'SetInput_1',
                                'vtkBandedPolyDataContourFilter': \
                                    'SetInput_1',
                                'vtkWarpScalar': 'SetInput_1'},
                   'SetSourceConnection': {None: 'SetSourceConnection',
                                           'vtkGlyph3D': \
                                               'SetSourceConnection_2'},
                   'AddFunction': {None: 'AddFunction',
                                   ('vtkImplicitSum', 'vtkPlane'): \
                                       'AddFunction_2',
                                   ('vtkImplicitSum', 'Tuple'): \
                                       'AddFunction_1'},
                   'SetColor': {None: 'SetColor',
                                ('vtkVolumeProperty', 
                                 'vtkColorTransferFunction'): \
                                    'SetColor_4',
                                ('vtkVolumeProperty', 
                                 'vtkPiecewiseFunction'): \
                                    'SetColor_2'},
                   'SetScalarOpacity': {None: 'SetScalarOpacity',
                                        'vtkVolumeProperty': \
                                            'SetScalarOpacity_2'},
                   'SetGradientOpacity': {None: 'SetGradientOpacity',
                                          'vtkVolumeProperty':
                                              'SetGradientOpacity_2'},
                   'SetSource': {None: 'SetSource',
                                 'vtkGlyph3D': 'SetSource_1'},
                   }

def translate_vtk(module_id, port_name, specs=None):
    global module_map
    if port_name in translate_ports:
        if module_id in module_map:
            (module_name, module_pkg) = module_map[module_id]
            if module_pkg != 'edu.utah.sci.vistrails.vtk':
                return port_name
        if module_name in translate_ports[port_name]:
            port_name = translate_ports[port_name][module_name]
        elif (module_name, specs) in translate_ports[port_name]:
            port_name = translate_ports[port_name][(module_name, specs)]
        else:
            port_name = translate_ports[port_name][None]  
    return port_name

def convert_data(child, parent_type, parent_id):
    from vistrails.core.vistrail.port_spec import PortSpec
    from vistrails.core.modules.module_registry import get_module_registry
    global module_map, translate_ports

    registry = get_module_registry()
    if child.vtType == 'module':
        descriptor = registry.get_descriptor_from_name_only(child.db_name)
        package = descriptor.identifier
        module_map[child.db_id] = (child.db_name, package)
        return DBModule(id=child.db_id,
                        cache=child.db_cache, 
                        abstraction=0, 
                        name=child.db_name,
                        package=package)
    elif child.vtType == 'connection':
        return DBConnection(id=child.db_id)
    elif child.vtType == 'portSpec':
        return DBPortSpec(id=child.db_id,
                          name=child.db_name, 
                          type=child.db_type, 
                          spec=child.db_spec)
    elif child.vtType == 'function':
        if parent_type == 'module':
            name = translate_vtk(parent_id, child.db_name)
        else:
            name = child.db_name
        return DBFunction(id=child.db_id,
                          pos=child.db_pos, 
                          name=name)
    elif child.vtType == 'parameter':
        return DBParameter(id=child.db_id,
                           pos=child.db_pos,
                           name=child.db_name, 
                           type=child.db_type, 
                           val=child.db_val, 
                           alias=child.db_alias)
    elif child.vtType == 'location':
        return DBLocation(id=child.db_id,
                          x=child.db_x,
                          y=child.db_y)
    elif child.vtType == 'annotation':
        return DBAnnotation(id=child.db_id,
                            key=child.db_key,
                            value=child.db_value)
    elif child.vtType == 'port':
        sig = child.db_sig
        if '(' in sig and ')' in sig:
            name = sig[:sig.find('(')]
            specs = sig[sig.find('(')+1:sig.find(')')]
            name = translate_vtk(child.db_moduleId, name, specs)

            new_specs = []
            for spec_name in specs.split(','):
                descriptor = registry.get_descriptor_from_name_only(spec_name)
                spec_str = descriptor.identifier + ':' + spec_name
                new_specs.append(spec_str)
            spec = '(' + ','.join(new_specs) + ')'
        else:
            name = sig
            spec = ''
        return DBPort(id=child.db_id,
                      type=child.db_type, 
                      moduleId=child.db_moduleId, 
                      moduleName=child.db_moduleName, 
                      name=name,
                      spec=spec)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.bundles.pyimport import py_import
import vistrails.core.requirements
from vistrails.gui.modules.source_configure import SourceConfigurationWidget
from PyQt4 import QtCore, QtGui
from vistrails.gui.theme import CurrentTheme

class PythonHighlighter(QtGui.QSyntaxHighlighter):
    def __init__( self, document ):
        QtGui.QSyntaxHighlighter.__init__( self, document )
        self.rules = []
        
        keywordFormat = QtGui.QTextCharFormat()
        keywordFormat.setForeground(QtCore.Qt.blue)
        self.rules += [(r"\b%s\b"%w, keywordFormat, -1, -1, False)
                       for w in ["def","from", 
                                 "import","for","in", 
                                 "while","True","None",
                                 "False","pass","return",
                                 "self","tuple","list",
                                 "print","if","else",
                                 "elif","len","string"
                                 "assert","try","except",
                                 "exec", "break", "continue",
                                 "not", "and", "or", "as",
                                 "type", "int", "float",
                                 ]]
        
        defclassFormat = QtGui.QTextCharFormat()
        defclassFormat.setForeground(QtCore.Qt.blue)
        self.rules += [(r"\bclass\b\s*(\w+)", defclassFormat, -1, -1, False)]

        commentFormat = QtGui.QTextCharFormat()
        commentFormat.setFontItalic(True)
        commentFormat.setForeground(QtCore.Qt.darkGreen)
        self.rules += [(r"#[^\n]*", commentFormat, -1, -1, False)]

        literalFormat = QtGui.QTextCharFormat()
        literalFormat.setForeground(QtGui.QColor(65, 105, 225)) #royalblue
        self.rules += [
            # Whole docstring
            (r"'''.*'''", literalFormat, -1, -1, True),
            (r'""".*"""', literalFormat, -1, -1, True),
            # Hanging docstring (single quote)
            (r"'''.*$", literalFormat, -1, 2, False),
            (r"^.*'''", literalFormat, 2, -1, True),
            (r"^.*$", literalFormat, 2, 2, False),
            # Hanging docstring (double quote)
            (r'""".*$', literalFormat, -1, 3, False),
            (r'^.*"""', literalFormat, 3, -1, True),
            (r'^.*$', literalFormat, 3, 3, False),
            # Whole string
            (r"'[^']*'", literalFormat, -1, -1, False),
            (r'"[^"]*"', literalFormat, -1, -1, False),
            # Hanging string (single quote)
            (r"'[^']*$", literalFormat, -1,  0, False),
            (r"^[^']*$", literalFormat,  0,  0, False),
            (r"^[^']*'", literalFormat,  0, -1, False),
            # Hanging string (double quotes)
            (r'"[^"]*$', literalFormat, -1,  1, False),
            (r'^[^"]*$', literalFormat,  1,  1, False),
            (r'^[^"]*"', literalFormat,  1, -1, False),
            ]
        
    def highlightBlock(self, text):
        baseFormat = self.format(0)
        prevState = self.previousBlockState()
        self.setCurrentBlockState(prevState)        
        shift = 0
        while True:
            matchedRule = (None, -1, -1, -1, -1)
            for rule in self.rules:
                if rule[2]==self.currentBlockState():
                    RE = QtCore.QRegExp(rule[0])
                    RE.setMinimal(rule[4])
                    pos = RE.indexIn(text, shift)
                    if (pos!=-1 and (matchedRule[0]==None or pos<matchedRule[1]) and
                        RE.matchedLength()>0):
                        matchedRule = (rule[1], pos, RE.matchedLength(), rule[3])
            if matchedRule[0]!=None:
                self.setFormat(matchedRule[1], matchedRule[2], matchedRule[0])
                self.setCurrentBlockState(matchedRule[3])
                shift = matchedRule[1]+matchedRule[2]
            else:
                break

def PythonEditor(parent=None):
    try:
        py_import('PyQt4.Qsci', {'linux-ubuntu': 'python-qscintilla2'})
    except ImportError:
        return OldPythonEditor(parent)
    else:
        return NewPythonEditor(parent)

def NewPythonEditor(parent):
    vistrails.core.requirements.require_python_module('PyQt4.Qsci')
    from PyQt4.Qsci import QsciScintilla, QsciLexerPython
    class _PythonEditor(QsciScintilla):
    
        def __init__(self, parent=None):
            QsciScintilla.__init__(self, parent)
            ## set the default font of the editor
            ## and take the same font for line numbers
            font = CurrentTheme.PYTHON_SOURCE_EDITOR_FONT
            self.setFont(font)
            fm = QtGui.QFontMetrics(font)
        
            ## Line numbers
            # conventionally, margin 0 is for line numbers
            self.setMarginWidth(0, fm.width( "0000" ) + 4)
            self.setMarginLineNumbers(0, True)
            self.setAutoIndent(True)

            ## Edge Mode shows a red vetical bar at 80 chars
            self.setEdgeMode(QsciScintilla.EdgeLine)
            self.setEdgeColumn(80)
            self.setEdgeColor(QtGui.QColor("#CCCCCC"))
        
            ## Folding visual : we will use boxes
            self.setFolding(QsciScintilla.BoxedTreeFoldStyle)
        
            ## Braces matching
            self.setBraceMatching(QsciScintilla.SloppyBraceMatch)
        
            ## Editing line color
        #    self.setCaretLineVisible(True)
        #    self.setCaretLineBackgroundColor(QtGui.QColor("#CDA869"))
            
            
                    ## Margins colors
            # line numbers margin
            self.setMarginsBackgroundColor(QtGui.QColor("#FFFFFF"))
            self.setMarginsForegroundColor(QtGui.QColor("#000000"))
        
            # folding margin colors (foreground,background)
            self.setFoldMarginColors(QtGui.QColor("#DDDDDD"),QtGui.QColor("#DDDDDD"))
            # do not use tabs
            self.setIndentationsUseTabs(False)
            self.setTabWidth(4)
            self.setTabIndents(True)
        
            ## Choose a lexer
            lexer = QsciLexerPython()
            lexer.setDefaultFont(font)
            lexer.setFont(font)
            self.setLexer(lexer)
        
            # set autocompletion
            self.setAutoCompletionThreshold(2)
            self.setAutoCompletionSource(QsciScintilla.AcsDocument)
            self.setAutoCompletionCaseSensitivity(True)
            self.setAutoCompletionReplaceWord(True)
            self.setAutoCompletionFillupsEnabled(True)
            
        def setPlainText(self, text):
            """ setPlainText(text: str) -> None
            redirect to setText
            
            """
            self.setText(text)
        
        def toPlainText(self):
            """ setPlainText(text: str) -> None
            redirect to self.text()
            
            """
            text = self.text()
            return text.replace('\r\n', '\n').replace('\r', '\n')
    
#        def focusOutEvent(self, event):
#            if self.parent():
#                QtCore.QCoreApplication.sendEvent(self.parent(), event)
#            QsciScintilla.focusOutEvent(self, event)

    return _PythonEditor(parent)

class OldPythonEditor(QtGui.QTextEdit):

    def __init__(self, parent=None):
        QtGui.QTextEdit.__init__(self, parent)
        self.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.formatChanged(None)
        self.setCursorWidth(8)
        self.highlighter = PythonHighlighter(self.document())
        self.connect(self,
                     QtCore.SIGNAL('currentCharFormatChanged(QTextCharFormat)'),
                     self.formatChanged)

    def formatChanged(self, f):
        self.setFont(CurrentTheme.PYTHON_SOURCE_EDITOR_FONT)

    def keyPressEvent(self, event):
        """ keyPressEvent(event: QKeyEvent) -> Nont
        Handle tab with 4 spaces
        
        """
        if event.key()==QtCore.Qt.Key_Tab:
            self.insertPlainText('    ')
        else:
            # super(PythonEditor, self).keyPressEvent(event)
            QtGui.QTextEdit.keyPressEvent(self, event)
            
#    def focusOutEvent(self, event):
#        if self.parent():
#            QtCore.QCoreApplication.sendEvent(self.parent(), event)
#        QtGui.QTextEdit.focusOutEvent(self, event)

class PythonSourceConfigurationWidget(SourceConfigurationWidget):
    def __init__(self, module, controller, parent=None):
        SourceConfigurationWidget.__init__(self, module, controller, 
                                           PythonEditor, True, True, parent)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.db.versions.v1_0_1.domain import DBVistrail, DBWorkflow, DBLog, \
    DBRegistry, DBGroup, DBTag, DBAnnotation, DBAction, IdScope

def translateVistrail(_vistrail):
    tag_annotations = {}
    notes_annotations = {}
    thumb_annotations = {}
    upgrade_annotations = {}
    prune_annotations = {}

    del_tag_annotations = {}
    del_notes_annotations = {}
    del_thumb_annotations = {}
    del_upgrade_annotations = {}
    del_prune_annotations = {}

    key_lists = {'__tag__': tag_annotations,
                 '__notes__': notes_annotations,
                 '__thumb__': thumb_annotations,
                 '__upgrade__': upgrade_annotations,
                 '__prune__': prune_annotations}

    del_key_lists = {'__tag__': del_tag_annotations,
                     '__notes__': del_notes_annotations,
                     '__thumb__': del_thumb_annotations,
                     '__upgrade__': del_upgrade_annotations,
                     '__prune__': del_prune_annotations}

    _vistrail.update_id_scope()
    id_scope = _vistrail.idScope

    def update_tags(old_obj, translate_dict):
        new_tags = []
        for (id, (_, tag, is_new, is_dirty)) in tag_annotations.iteritems():
            new_tag = DBTag(id=id, name=tag)
            new_tag.is_new = is_new
            new_tag.is_dirty = is_dirty
            new_tags.append(new_tag)
        return new_tags

    def update_prune(old_obj, translate_dict):
        if old_obj.db_id in prune_annotations:
            (_, prune_val, _, _) = prune_annotations[old_obj.db_id]
            if prune_val == str(True):
                return 1
            elif prune_val == str(False):
                return 0
        return None

    def update_annotations(old_obj, translate_dict):
        new_annotations = [DBAnnotation.update_version(a, translate_dict)
                           for a in old_obj.db_annotations]
        if old_obj.db_id in notes_annotations:
            (id, notes, is_new, is_dirty) = notes_annotations[old_obj.db_id]
            ann = DBAnnotation(id=id,
                               key='__notes__',
                               value=notes)
            ann.is_new = is_new
            ann.is_dirty = is_dirty
            new_annotations.append(ann)
        if old_obj.db_id in upgrade_annotations:
            (id, upgrade, is_new, is_dirty) = \
                upgrade_annotations[old_obj.db_id]
            ann =  DBAnnotation(id=id,
                                key='__upgrade__',
                                value=upgrade)
            ann.is_new = is_new
            ann.is_dirty = is_dirty
            new_annotations.append(ann)
        if old_obj.db_id in thumb_annotations:
            (id, thumb, is_new, is_dirty) = thumb_annotations[old_obj.db_id]
            ann = DBAnnotation(id=id,
                               key='__thumb__',
                               value=thumb)
            ann.is_new = is_new
            ann.is_dirty = is_dirty
            new_annotations.append(ann)
        return new_annotations

    def update_actions(old_obj, translate_dict):
        new_actions = []
        for action in old_obj.db_actions:
            if action.db_id in del_notes_annotations:
                (id, notes, is_new, is_dirty) = \
                    del_notes_annotations[action.db_id]
                ann = DBAnnotation(id=id,
                                   key='__notes__',
                                   value=notes)
                ann.is_new = is_new
                ann.is_dirty = is_dirty
                action.db_deleted_annotations.append(ann)
            if action.db_id in del_upgrade_annotations:
                (id, upgrade, is_new, is_dirty) = \
                    del_upgrade_annotations[action.db_id]
                ann = DBAnnotation(id=id,
                                   key='__upgrade__',
                                   value=upgrade)
                ann.is_new = is_new
                ann.is_dirty = is_dirty
                action.db_deleted_annotations.append(ann)
            if action.db_id in del_thumb_annotations:
                (id, thumb, is_new, is_dirty) = \
                    del_thumb_annotations[action.db_id]
                ann = DBAnnotation(id=id,
                                   key='__thumb__',
                                   value=thumb)
                ann.is_new = is_new
                ann.is_dirty = is_dirty
                action.db_deleted_annotations.append(ann)                
            new_actions.append(DBAction.update_version(action, translate_dict))
        return new_actions

    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)

    for a in _vistrail.db_actionAnnotations:
        if a.db_key in key_lists:
            key_lists[a.db_key][a.db_action_id] = \
                (a.db_id, a.db_value, a.is_new, a.is_dirty)
    for a in _vistrail.db_deleted_actionAnnotations:
        if a.db_key in del_key_lists:
            del_key_lists[a.db_key][a.db_action_id] = \
                (a.db_id, a.db_value, a.is_new, a.is_dirty)

    translate_dict = {'DBGroup': {'workflow': update_workflow},
                      'DBVistrail': {'tags': update_tags,
                                     'actions': update_actions},
                      'DBAction': {'annotations': update_annotations,
                                   'prune': update_prune}}
    vistrail = DBVistrail.update_version(_vistrail, translate_dict)
    for (id, (_, tag, is_new, is_dirty)) in del_tag_annotations.iteritems():
        new_tag = DBTag(id=id, name=tag)
        new_tag.is_new = is_new
        new_tag.is_dirty = is_dirty
        vistrail.db_deleted_tags.append(new_tag)
    
    vistrail.db_version = '1.0.1'
    return vistrail

def translateWorkflow(_workflow):
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBGroup': {'workflow': update_workflow}}
    workflow = DBWorkflow.update_version(_workflow, translate_dict)
    workflow.db_version = '1.0.1'
    return workflow

def translateLog(_log):
    translate_dict = {}
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '1.0.1'
    return log

def translateRegistry(_registry):
    translate_dict = {}
    registry = DBRegistry.update_version(_registry, translate_dict)
    registry.db_version = '1.0.1'
    return registry



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from itertools import izip
import re

def parse_vt_command(opt_arg, arg):
    opt_dict = {}
    for opt_str in opt_arg[1:-1].split(','):
        opt_str = opt_str.strip()
        opt_list = opt_str.split('=',1)
        if len(opt_list) > 1:
            if opt_list[0].strip():
                opt_dict[opt_list[0]] = opt_list[1]
        else:
            if opt_list[0].strip():
                opt_dict[opt_list[0]] = None
    opt_dict['_args'] = arg[1:-1]
    return opt_dict

def build_vt_command(opt_dict):
    if '_args' in opt_dict:
        args = opt_dict['_args']
    else:
        args = ''
    opt_str = ''
    for k, v in opt_dict.iteritems():
        if k == '_args':
            continue
        if v is None:
            opt_str += '%s,\n' % k
        else:
            opt_str += '%s=%s,\n' % (k,v)
    if opt_str:
        full_str = "\\vistrail[%s]{%s}" % (opt_str, args)
    else:
        full_str = "\\vistrail{%s}" % args
    #print '^^ LATEX ^^', opt_dict, full_str
    return full_str

def parse_latex_file(fname):
    # vt_begin_re = re.compile(r"% -- VISTRAILS BEGIN --")
    # vt_end_re = re.compile(r"% -- VISTRAILS END --")
    vt_begin_re = re.compile(r"^(?:(?:\\%)|[^%])*(\\vistrail[\[\{])")

    # we'll just assume balanced parens here...
    # vt_end_re = re.compile(r"\}")
    vt_bracket_re = re.compile(r"[\[\]]")
    vt_brace_re = re.compile(r"[\{\}]")
    f = open(fname)
    
    raw_text = [""]
    raw_idx = 0
    vt_text = []
    for line in f:
        m = vt_begin_re.search(line)
        if not m:
            raw_text[raw_idx] += line
        else:
            # print 'found:', m.start(1), m.end(1), line[m.end(1) - 1]

            # first end the current raw_text block
            raw_text[raw_idx] += line[:m.start(1)]
            raw_text.append("")
            raw_idx += 1

            cmd_text = ""
            # line = line[m.start(1):]
            line = line[m.end(1)-1:]

            # we'll just assume balanced parens here...
            if line[0] == '[':
                brackets_complete = False
                found_start = False
                brackets = 0
                while not brackets_complete:
                    for m in vt_bracket_re.finditer(line):
                        bracket_char = line[m.start()]
                        if bracket_char == '[':
                            brackets += 1
                            found_start = True
                        else:
                            brackets -= 1
                        if found_start and brackets == 0:
                            brackets_complete = True
                            cmd_text += line[:m.end()]
                            line = line[m.end():]
                            break
                    if brackets_complete:
                        break
                    cmd_text += line
                    line = f.next()
            
            opt_cmd_text = cmd_text
            cmd_text = ""
            braces_complete = False
            found_start = False
            braces = 0
            while not braces_complete:
                for m in vt_brace_re.finditer(line):
                    brace_char = line[m.start()]
                    if brace_char == '{':
                        braces += 1
                        found_start = True
                    else:
                        braces -= 1
                    if found_start and braces == 0:
                        braces_complete = True
                        cmd_text += line[:m.end()]
                        break
                if braces_complete:
                    break
                cmd_text += line
                line = f.next()
            # print cmd_text
            vt_text.append((opt_cmd_text,cmd_text))
            raw_text[raw_idx] += line[m.end():]
            
        # if vt_begin_re.search(line):
        #     line = f.next()
        #     while not vt_end_re.search(line):
        #         print line.strip()
        #         line = f.next()
    f.close()

    # vt_text.append(("", ""))
    #for t, vt in izip(raw_text, vt_text + [("", "")]):
        #print "RAW:", t
        #print "VT:", parse_vt_command(*vt)
    return (raw_text, vt_text)

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        print "Usage: %s %s <latex-file>" % (sys.executable, sys.argv[0])
        sys.exit(-1)
    
    parse_latex_file(sys.argv[1])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.domain import IdScope
from vistrails.db.domain import DBMashup
from vistrails.core.mashup import conv_to_bool, conv_from_bool
from vistrails.core.mashup.alias import Alias
from vistrails.core.mashup.component import Component
from vistrails.core.system import get_elementtree_library

import unittest
import copy

ElementTree = get_elementtree_library()

class Mashup(DBMashup):
    def __init__(self, id, name, vtid=None, version=None, alias_list=None, 
                 t='vistrail', has_seq=None, layout='', geometry='', 
                 id_scope=IdScope()):
        if has_seq is None:
            has_seq = 0
            
        DBMashup.__init__(self, id, name, version, alias_list, t, vtid, layout, 
                          geometry, has_seq)
        self.id_scope = id_scope
        if has_seq is None:
            self.has_seq = False
            if isinstance(self.alias_list, list):
                for v in self.alias_list:
                    if v.component.seq == True:
                        self.has_seq = True
        else:
            self.has_seq = has_seq
    id = DBMashup.db_id
    name = DBMashup.db_name
    version = DBMashup.db_version
    alias_list = DBMashup.db_aliases
    aliases = DBMashup.db_aliases
    vtid = DBMashup.db_vtid
    type = DBMashup.db_type
    layout = DBMashup.db_layout
    geometry = DBMashup.db_geometry
        
    def _get_has_seq(self):
        return conv_to_bool(self.db_has_seq)
    def _set_has_seq(self, s):
        self.db_has_seq = conv_from_bool(s)
    has_seq = property(_get_has_seq,_set_has_seq)
    
    @staticmethod
    def convert(_mashup):
        _mashup.__class__ = Mashup

        for alias in _mashup.alias_list:
            Alias.convert(alias)
            
    def __copy__(self):
        return Mashup.do_copy(self)
    
    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        """do_copy() -> Mashup 
        returns a clone of itself"""
        cp = DBMashup.do_copy(self, new_ids, id_scope, id_remap)
        Mashup.convert(cp)
        return cp
    
    ##########################################################################
    # Serialization / Unserialization
    
#    def toXml(self, node=None):
#        """toXml(node: ElementTree.Element) -> ElementTree.Element
#           writes itself to xml
#        """
#
#        if node is None:
#            node = ElementTree.Element('mashup')
#
#        #set attributes
#        node.set('id', self.convert_to_str(self.id,'long'))
#        node.set('version', self.convert_to_str(self.version,'long'))
#        node.set('vtid', self.convert_to_str(self.vtid,'str'))
#        node.set('name', self.convert_to_str(self.name,'str'))
#        node.set('type', self.convert_to_str(self.type,'str'))
#        node.set('has_seq', self.convert_to_str(self.has_seq,'bool'))
#        for v in self.alias_list:
#            child_ = ElementTree.SubElement(node, 'alias')
#            v.toXml(child_)
#        
#        layoutnode = ElementTree.SubElement(node,'layout')
#        layoutnode.text = str(self.layout)
#        
#        geomnode = ElementTree.SubElement(node,'geometry')
#        geomnode.text = str(self.geometry)
#        
#        return node
#
#    @staticmethod
#    def fromXml(node):
#        if node.tag != 'mashup':
#            #print "node.tag != 'mashup'"
#            return None
#        #read attributes
#        data = node.get('id', None)
#        id = Mashup.convert_from_str(data, 'long')
#        data = node.get('name', None)
#        name = Mashup.convert_from_str(data, 'str')
#        data = node.get('version', None)
#        version = Mashup.convert_from_str(data, 'long')
#        data = node.get('vtid', None)
#        vtid = Mashup.convert_from_str(data, 'str')
#        data = node.get('type', None)
#        type = Mashup.convert_from_str(data, 'str')
#        data = node.get('has_seq', None)
#        seq = Component.convert_from_str(data, 'bool')
#        alias_list = []
#        layout = None
#        geometry = None
#        for child in node.getchildren():
#            if child.tag == "alias":
#                alias = Alias.fromXml(child)
#                alias_list.append(alias)
#            if child.tag == "layout":
#                layout = str(child.text).strip(" \n\t")
#            if child.tag == "geometry":
#                geometry = str(child.text).strip(" \n\t")
#        return Mashup(id=id, name=name, vtid=vtid, version=version, 
#                               alias_list=alias_list, t=type, has_seq=seq,
#                               layout=layout, geometry=geometry)
        
    def loadAliasesFromPipeline(self, pipeline, id_scope):
        """loadAliasesFromPipelines(pipeline: Pipeline) -> None 
        This will replace current aliases with the ones present in Pipeline
        
        """
        
        if pipeline:
            self.alias_list = []
            if len(pipeline.aliases) > 0:
                pos = 0
                for aname,info in pipeline.aliases.iteritems():
                    parameter = pipeline.db_get_object(info[0],info[1])
                    cid = id_scope.getNewId('component')
                    aid = id_scope.getNewId('alias')
                    component = Component(cid, parameter.vtType, 
                                          parameter.real_id, info[2], info[3],
                                          info[4], parameter.type, 
                                          parameter.strValue, parameter.pos, 
                                          pos, "")
                    alias = Alias(aid, aname, component)
                    self.alias_list.append(alias)
                    pos += 1 
                  
    def remapPipelineObjects(self, id_remap):
        for alias in self.alias_list:
            try:
                new_pid = id_remap[(alias.component.vtparent_type,
                                    alias.component.vtparent_id)]
                alias.component.vtparent_id = new_pid
                new_id = id_remap[(alias.component.vttype,alias.component.vtid)]
                alias.component.vtid = new_id
            except Exception:
                pass

    def validateForPipeline(self, pipeline):
        """validateForPipeline(pipeline) -> None
        This will make sure that the parameters in the alias list are present
        in the pipeline. If they were removed, the aliases pointing to it will 
        be removed. This changes the mashup in place """
        to_remove = []
        for alias in self.alias_list:
            try:
                param = pipeline.db_get_object(alias.component.vttype,
                                               alias.component.vtid)
            except Exception:
                to_remove.append(alias)
        for a in to_remove:
            self.alias_list.remove(a)    
        pos = 0
        mashup_aliases = []
        for a in self.alias_list:
            mashup_aliases.append(a.name)
            a.component.pos = pos
            pos+=1
            
        for a, info in pipeline.aliases.iteritems():
            if a not in mashup_aliases:
                parameter = pipeline.db_get_object(info[0],info[1])
                cid = self.id_scope.getNewId('component')
                aid = self.id_scope.getNewId('alias')
                component = Component(cid, parameter.vtType, 
                                      parameter.real_id, info[2], info[3],
                                      info[4], parameter.type, 
                                      parameter.strValue, parameter.pos, 
                                      pos, "")
                newalias = Alias(aid, a, component) 
                self.alias_list.append(newalias)
                pos +=1      
                   
    def getAliasByName(self, name):
        for alias in self.alias_list:
            if alias.name == name:
                return alias
                        
    ##########################################################################
    # Operators

    def __str__(self):
        """ __str__() -> str - Returns a string representation of itself """
        
        return ("(Mashup id='%s' name='%s' version='%s' vtid='%s' type='%s' \
layout='%s' geometry='%s' alias_list='%s')@%X" %
                    (self.id,
                     self.name,
                     self.version,
                     self.vtid,
                     self.type,
                     self.layout,
                     self.geometry,
                     self.alias_list,
                     id(self)))

    def __eq__(self, other):
        """ __eq__(other: Mashup) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.name != other.name:
            return False
        if self.vtid != other.vtid:
            return False
        if self.version != other.version:
            return False
        if self.type != other.type:
            return False
        if self.layout != other.layout:
            return False
        if self.geometry != other.geometry:
            return False
        if len(self.alias_list) != len(other.alias_list):
            return False
        for p,q in zip(self.alias_list, other.alias_list):
            if p != q:
                return False
        return True

    def __ne__(self, other):
        """ __ne__(other: Component) -> boolean
        Returns True if self and other don't have the same attributes. 
        Used by !=  operator. 
        
        """
        return not self.__eq__(other)              
        
################################################################################


class TestMashup(unittest.TestCase):
    def create_mashup(self, id_scope=IdScope()):
        c1 = Component(id=id_scope.getNewId('mashup_component'),
                          vttype='parameter', param_id=15L, 
                          parent_vttype='function', parent_id=3L, mid=4L,
                          type='String', value='test', p_pos=0, pos=1, 
                          strvaluelist='test1,test2', widget="text")
        a1 = Alias(id=id_scope.getNewId('mashup_alias'), name='alias1', component=c1)
        
        m = Mashup(id=id_scope.getNewId('mashup'), name='mashup1', vtid='empty.vt', 
                   version=15L, alias_list=[a1])
        return m
    
    def test_copy(self):
        id_scope = IdScope()
        m1 = self.create_mashup(id_scope)
        m2 = copy.copy(m1)
        self.assertEqual(m1, m2)
        self.assertEqual(m1.id, m2.id)
        m3 = m2.do_copy(True, id_scope, {})
        self.assertEqual(m1, m3)
        self.assertNotEqual(m1.id, m3.id)
        
#    def test_serialization(self):
#        m1 = self.create_mashup()
#        node = m1.toXml()
#        m2 = Mashup.fromXml(node)
#        self.assertEqual(m1, m2)
#        self.assertEqual(m1.id, m2.id)
        
    def test_str(self):
        m1 = self.create_mashup()
        str(m1)
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Module with utilities to try and install a bundle if possible."""
from __future__ import division

from vistrails.core import get_vistrails_application
from vistrails.core.configuration import get_vistrails_configuration, \
    get_vistrails_persistent_configuration
from vistrails.core import debug
from vistrails.core.system import executable_is_in_path, get_executable_path
from vistrails.core.system import vistrails_root_directory, systemType
from vistrails.gui.bundles.utils import guess_system, guess_graphical_sudo
import vistrails.gui.bundles.installbundle # this is on purpose
from vistrails.gui.requirements import qt_available
import imp
import os
import subprocess
import sys

##############################################################################

pip_installed = True
try:
    imp.find_module('pip')
    # Here we do not actually import pip, to avoid pip issue #1314
    # https://github.com/pypa/pip/issues/1314
except ImportError:
    pip_installed = False

def hide_splash_if_necessary():
    """Disables the splashscreen, otherwise it sits in front of windows.
    """
    app = get_vistrails_application()
    if hasattr(app, 'splashScreen') and app.splashScreen:
        app.splashScreen.hide()


def shell_escape(arg):
    return '"%s"' % arg.replace('\\', '\\\\').replace('"', '\\"')


def run_install_command(graphical, cmd, args, as_root=True):
    if isinstance(args, str):
        cmd += ' %s' % shell_escape(args)
    elif isinstance(args, list):
        for package in args:
            if not isinstance(package, str):
                raise TypeError("Expected string or list of strings")
            cmd += ' %s' % shell_escape(package)
    else:
        raise TypeError("Expected string or list of strings")

    debug.warning("VisTrails wants to install package(s) %r" %
                  args)

    if as_root and systemType != 'Windows':
        if graphical:
            sucmd, escape = guess_graphical_sudo()
        else:
            if get_executable_path('sudo'):
                sucmd, escape = "sudo %s", False
            elif systemType != 'Darwin':
                sucmd, escape = "su -c %s", True
            else:
                sucmd, escape = '%s', False

        if escape:
            cmd = sucmd % shell_escape(cmd)
        else:
            cmd = sucmd % cmd

    print "about to run: %s" % cmd
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT,
                              shell=True)
    lines = []
    try:
        for line in iter(p.stdout.readline, ''):
            print line,
            lines.append(line)
    except IOError, e:
        print "Ignoring IOError:", str(e)
    result = p.wait()

    if result != 0:
        debug.critical("Error running: %s" % cmd, ''.join(lines))

    return result == 0 # 0 indicates success


def linux_debian_install(package_name):
    qt = qt_available()
    try:
        import apt
        import apt_pkg
    except ImportError:
        qt = False
    hide_splash_if_necessary()

    if qt:
        cmd = shell_escape(vistrails_root_directory() +
                           '/gui/bundles/linux_debian_install.py')
    else:
        cmd = '%s install -y' % ('aptitude'
                                 if executable_is_in_path('aptitude')
                                 else 'apt-get')

    return run_install_command(qt, cmd, package_name)

linux_ubuntu_install = linux_debian_install


def linux_fedora_install(package_name):
    qt = qt_available()
    hide_splash_if_necessary()

    if qt:
        cmd = shell_escape(vistrails_root_directory() +
                           '/gui/bundles/linux_fedora_install.py')
    else:
        cmd = 'yum -y install'

    return run_install_command(qt, cmd, package_name)


def pip_install(package_name):
    hide_splash_if_necessary()

    if executable_is_in_path('pip'):
        cmd = '%s install' % shell_escape(get_executable_path('pip'))
    else:
        cmd = shell_escape(sys.executable) + ' -m pip install'

    if systemType != 'Windows':
        use_root = True
        try:
            from distutils.sysconfig import get_python_lib
            f = get_python_lib()
        except Exception:
            f = sys.executable
        use_root = os.stat(f).st_uid == 0
    else:
        use_root = False

    return run_install_command(qt_available(), cmd, package_name, use_root)

def show_question(which_files, has_distro_pkg, has_pip):
    if isinstance(which_files, str):
        which_files = [which_files]
    if qt_available():
        from PyQt4 import QtCore, QtGui
        dialog = QtGui.QDialog()
        dialog.setWindowTitle("Required packages missing")
        layout = QtGui.QVBoxLayout()

        label = QtGui.QLabel(
                "One or more required packages are missing: %s. VisTrails can "
                "automatically install them. If you click OK, VisTrails will "
                "need administrator privileges, and you might be asked for "
                "the administrator password." % (" ".join(which_files)))
        label.setWordWrap(True)
        layout.addWidget(label)

        if pip_installed and has_pip:
            use_pip = QtGui.QCheckBox("Use pip")
            use_pip.setChecked(
                not has_distro_pkg or (
                    has_pip and
                    getattr(get_vistrails_configuration(),
                            'installBundlesWithPip')))
            use_pip.setEnabled(has_distro_pkg and has_pip)
            layout.addWidget(use_pip)

            remember_align = QtGui.QHBoxLayout()
            remember_align.addSpacing(20)
            remember_pip = QtGui.QCheckBox("Remember my choice")
            remember_pip.setChecked(False)
            remember_pip.setEnabled(use_pip.isEnabled())
            remember_align.addWidget(remember_pip)
            layout.addLayout(remember_align)
        elif has_pip:
            label = QtGui.QLabel("pip package is available but pip is not installed")
            layout.addWidget(label)
        buttons = QtGui.QDialogButtonBox(
                QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel)
        QtCore.QObject.connect(buttons, QtCore.SIGNAL('accepted()'),
                               dialog, QtCore.SLOT('accept()'))
        QtCore.QObject.connect(buttons, QtCore.SIGNAL('rejected()'),
                               dialog, QtCore.SLOT('reject()'))
        layout.addWidget(buttons)

        dialog.setLayout(layout)
        hide_splash_if_necessary()
        if dialog.exec_() != QtGui.QDialog.Accepted:
            return False
        else:
            if pip_installed and has_pip:
                if remember_pip.isChecked():
                    setattr(get_vistrails_persistent_configuration(),
                            'installBundlesWithPip',
                            use_pip.isChecked())

                if use_pip.isChecked():
                    return 'pip'
            return 'distro'
    else:
        print "\nRequired package(s) missing: %s" % (" ".join(which_files))
        print ("A required package is missing, but VisTrails can "
               "automatically install it. "
               "If you say Yes, VisTrails will need "
               "administrator privileges, and you "
               "might be asked for the administrator password.")
        if has_distro_pkg:
            print "(VisTrails will use your distribution's package manager)"
        else:
            print "(VisTrails will use the 'pip' installer)"
        print "Give VisTrails permission to try to install package? (y/N)"
        v = raw_input().upper()
        if v == 'Y' or v == 'YES':
            if has_distro_pkg:
                return 'distro'
            else:
                return 'pip'


def install(dependency_dictionary):
    """Tries to install a bundle after a py_import() failed.."""

    distro = guess_system()
    files = (dependency_dictionary.get(distro) or
             dependency_dictionary.get('pip'))
    if not files:
        return None
    can_install = (('pip' in dependency_dictionary and pip_installed) or
                   distro in dependency_dictionary)
    if can_install:
        action = show_question(
                files,
                distro in dependency_dictionary,
                'pip' in dependency_dictionary)
        if action == 'distro':
            callable_ = getattr(vistrails.gui.bundles.installbundle,
                                distro.replace('-', '_') + '_install')
            return callable_(files)
        elif action == 'pip':
            if not pip_installed:
                debug.warning("Attempted to use pip, but it is not installed.")
                return False
            return pip_install(dependency_dictionary.get('pip'))
        else:
            return False



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""This file specifies the configuration widget for OutputModule and
its subclasses.

"""

from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.configuration import ConfigPath, ConfigField
from vistrails.core.modules.basic_modules import Dictionary
from vistrails.gui.common_widgets import QDirectoryChooserToolButton
from vistrails.gui.modules.module_configure import StandardModuleConfigurationWidget

class OutputModuleConfigurationWidget(StandardModuleConfigurationWidget):
    def __init__(self, module, controller, parent=None):
        StandardModuleConfigurationWidget.__init__(self, module, controller, 
                                                   parent)
        self.update_widget()

    def get_configuration(self):
        config = {}
        for f in self.module.functions:
            if f.name == 'configuration':
                f_config = Dictionary.translate_to_python(f.params[0].strValue)
                config.update(f_config)
        return config

    def set_configuration(self, config):
        self.controller.update_function(self.module, 'configuration', 
                                [Dictionary.translate_to_string(config)])

    def update_widget(self):
        layout = QtGui.QVBoxLayout()
        config = self.get_configuration()

        self.mode_widgets = []
        self.found_modes = set()
        mode_layouts = []
        for mode in self.module.module_descriptor.module.get_sorted_mode_list():
            mode_config = None
            if mode.mode_type in config:
                mode_config = config[mode.mode_type]
            # create output mode widget passing current config
            mode_w = self.build_mode_config(layout, mode, mode_config)
            mode_layouts.append(mode_w.layout())
            self.found_modes.add(mode.mode_type)
            
        for mode_type, mode_config in config.iteritems():
            if mode_type not in self.found_modes:
                mode_w = self.build_mode_config(layout, None, mode_config, 
                                                title=mode_type)
                mode_layouts.append(mode_w.layout())
                self.found_modes.add(mode_type)
                
        width = 0
        for mode_layout in mode_layouts:
            for row in xrange(mode_layout.rowCount()):
                item = mode_layout.itemAtPosition(row, 0)
                if item and item.widget():
                    width = max(width, item.widget().sizeHint().width())
        for mode_layout in mode_layouts:
            mode_layout.setColumnMinimumWidth(0, width)

        # do we want to add a manual config mode for modes that have
        # neither been set before nor are registered?
        # DK: not now...
        layout.addStretch(5)
        scroll_area = QtGui.QScrollArea()
        inner_widget =  QtGui.QWidget()
        inner_widget.setLayout(layout)
        scroll_area.setWidget(inner_widget)
        scroll_area.setWidgetResizable(True)
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().addWidget(scroll_area, 1)
        self.layout().addLayout(self.create_buttons())
        self.layout().setContentsMargins(0,0,0,0)

    def build_mode_config(self, base_layout, mode, mode_config, title=None):
        mode_widget = OutputModeConfigurationWidget(mode, mode_config, title)
        mode_widget.fieldChanged.connect(self.field_was_changed)
        self.mode_widgets.append(mode_widget)
        base_layout.addWidget(mode_widget, 1)
        return mode_widget

    def create_buttons(self):
        """ create_buttons() -> None
        Create and connect signals to Save & Reset button
        
        """
        buttonLayout = QtGui.QHBoxLayout()
        buttonLayout.setMargin(5)
        self.saveButton = QtGui.QPushButton('&Save', self)
        self.saveButton.setFixedWidth(100)
        self.saveButton.setEnabled(False)
        buttonLayout.addWidget(self.saveButton)
        self.resetButton = QtGui.QPushButton('&Reset', self)
        self.resetButton.setFixedWidth(100)
        self.resetButton.setEnabled(False)
        buttonLayout.addWidget(self.resetButton)
        self.connect(self.saveButton, QtCore.SIGNAL('clicked(bool)'),
                     self.save_triggered)
        self.connect(self.resetButton, QtCore.SIGNAL('clicked(bool)'),
                     self.reset_triggered)
        return buttonLayout

    def save_triggered(self):
        # get values from each widget check if any changed and
        # then dump the whole dictionary back to the configuration
        # function
        config = self.get_configuration()
        for mode_widget in self.mode_widgets:
            config.update(mode_widget._changed_config)
        self.set_configuration(config)
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)

    def reset_triggered(self):
        config = self.get_configuration()
        for mode_widget in self.mode_widgets:
            for config_key, field in mode_widget._changed_fields.iteritems():
                widget = mode_widget.field_widgets[config_key]
                mode_type = config_key[0]
                mode_config = None
                if mode_type in config:
                    mode_config = config[mode_type]
                mode_widget.reset_field(widget, field, mode_config, mode_type)
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)

    def field_was_changed(self, mode_widget):
        self.saveButton.setEnabled(True)
        self.resetButton.setEnabled(True)

class OutputModeConfigurationWidget(QtGui.QGroupBox):
    fieldChanged = QtCore.pyqtSignal(object)

    def __init__(self, mode, mode_config=None, title=None, parent=None):
        assert(mode is not None or mode_config is not None)

        QtGui.QGroupBox.__init__(self, parent)
        self.field_widgets = {}
        self._changed_config = {}
        self._changed_fields = {}

        if title is not None:
            self.setTitle(title)
        elif mode is not None:
            self.setTitle(mode.mode_type)
        else:
            self.setTitle("unknown")

        group_layout = QtGui.QGridLayout()
        group_layout.setMargin(5)
        group_layout.setSpacing(5)
        group_layout.setColumnStretch(1,1)

        if mode is None:
            for k, v in mode_config.iteritems():
                dummy_field = ConfigField(k, None, str)
                self.add_field(group_layout, dummy_field, mode_config,
                                        k)
        else:
            for field in mode.get_config().get_all_fields():
                self.add_field(group_layout, field, mode_config, 
                               mode.mode_type)
        self.setLayout(group_layout)

    # TODO Unify this with code in gui.configuration!
    def add_field(self, layout, field, mode_config, mode_type, indent=0):
        config_key = (mode_type, field.name)
        if mode_config is not None and field.name in mode_config:
            config_val = mode_config[field.name]
        else:
            config_val = field.default_val

        config_desc = field.name
        widget_type = field.widget_type
        if widget_type is None:
            if field.val_type == bool:
                widget_type = "checkbox"
            elif field.val_type == ConfigPath:
                widget_type = "pathedit"
            else:
                widget_type = "lineedit"

        if widget_type == "combo":
            widget = self.add_combo(layout, field, config_key, config_desc, 
                                    config_val)
        elif widget_type == "lineedit":
            widget = self.add_line_edit(layout, field, config_key, config_desc, 
                               config_val)
        elif widget_type == "pathedit":
            widget = self.add_path_edit(layout, field, config_key, config_desc, 
                                        config_val)
        else:
            config_val = bool(config_val)
            widget = self.add_checkbox(layout, field, config_key, config_desc, 
                                       config_val)
        self.field_widgets[config_key] = widget

    def reset_field(self, widget, field, mode_config, mode_type):
        if mode_config is not None and field.name in mode_config:
            config_val = mode_config[field.name]
        else:
            config_val = field.default_val

        if field.widget_type == "checkbox":
            config_val = bool(config_val)
        self.set_value(widget, field, config_val)

    def set_value(self, widget, field, val):
        widget_type = field.widget_type
        if widget_type is None:
            if field.val_type == bool:
                widget_type = "checkbox"
            elif field.val_type == ConfigPath:
                widget_type = "pathedit"
            else:
                widget_type = "lineedit"

        if widget_type == "combo":
            self.set_combo_value(widget, val, field)
        elif widget_type == "lineedit":
            self.set_line_edit_value(widget, val)
        elif widget_type == "pathedit":
            self.set_path_edit_value(widget, val)
        else:
            self.set_checkbox_value(widget, val)

    def add_checkbox(self, layout, field, config_key, config_desc, config_val):
        cb = QtGui.QCheckBox(config_desc)
        self.set_checkbox_value(cb, config_val)
        row = layout.rowCount()
        layout.addWidget(cb, row, 1)

        def call_field_changed(val):
            self.field_changed(config_key, field, val, config_val)
        cb.toggled.connect(call_field_changed)
        return cb

    def set_checkbox_value(self, cb, config_val):
        cb.setChecked(config_val)

    def add_line_edit(self, layout, field, config_key, config_desc, config_val):
        options = {}
        if field.widget_options is not None:
            options = field.widget_options

        if "label" in options:
            label_text = options["label"]
        else:
            label_text = config_desc
        label = QtGui.QLabel(label_text)
        row = layout.rowCount()
        layout.addWidget(label, row, 0, QtCore.Qt.AlignRight)

        line_edit = QtGui.QLineEdit()
        self.set_line_edit_value(line_edit, config_val)
        layout.addWidget(line_edit, row, 1)

        def call_field_changed():
            val = line_edit.text()
            self.field_changed(config_key, field, val, config_val)
        line_edit.editingFinished.connect(call_field_changed)
        return line_edit

    def set_line_edit_value(self, line_edit, config_val):
        if config_val is None:
            config_val = ""
        line_edit.setText(unicode(config_val))

    def add_path_edit(self, layout, field, config_key, config_desc, config_val):
        options = {}
        if field.widget_options is not None:
            options = field.widget_options

        path_edit = QtGui.QWidget()
        if "label" in options:
            label_text = options["label"]
        else:
            label_text = config_desc
        label = QtGui.QLabel(label_text)
        row = layout.rowCount()
        layout.addWidget(label, row, 0, QtCore.Qt.AlignRight)

        sub_layout = QtGui.QHBoxLayout()
        sub_layout.setMargin(0)
        sub_layout.setSpacing(5)
        line_edit = QtGui.QLineEdit()
        if config_val is None:
            config_val = ""
        line_edit.setText(unicode(config_val))
        sub_layout.addWidget(line_edit)
        path_edit.line_edit = line_edit

        # if field.val_type == ConfigPath:
        #     button_cls = QDirectoryChooserToolButton
        # else:
        button_cls = QDirectoryChooserToolButton
        button = button_cls(self, line_edit)
        sub_layout.addWidget(button)
        path_edit.setLayout(sub_layout)
        layout.addWidget(path_edit, row, 1)

        def call_field_changed():
            val = line_edit.text()
            self.field_changed(config_key, field, val, config_val)
        line_edit.editingFinished.connect(call_field_changed)
        return path_edit

    def set_path_edit_value(self, path_edit, config_val):
        if config_val is None:
            config_val = ""
        path_edit.line_edit.setText(unicode(config_val))

    def add_combo(self, layout, field, config_key, config_desc, config_val):
        options = {}
        if field.widget_options is not None:
            options = field.widget_options

        if "label" in options:
            label_text = options["label"]
        else:
            label_text = config_desc
        label = QtGui.QLabel(label_text)
        row = layout.rowCount()
        layout.addWidget(label, row, 0, QtCore.Qt.AlignRight)

        combo = QtGui.QComboBox()
        inv_remap = None
        if "allowed_values" in options:
            values = options["allowed_values"]
            if "remap" in options:
                remap = options["remap"]
                inv_remap = dict((v, k) for (k, v) in remap.iteritems())
                entries = [remap[v] for v in values]
            else:
                entries = values
            for entry in entries:
                combo.addItem(entry)
        self.set_combo_value(combo, config_val, field)
        layout.addWidget(combo, row, 1)

        def call_field_changed(val):
            if inv_remap is not None:
                val = inv_remap[val]
            self.field_changed(config_key, field, val, config_val)
        combo.currentIndexChanged[unicode].connect(call_field_changed)
        return combo

    def set_combo_value(self, combo, config_val, field):
        options = {}
        if field.widget_options is not None:
            options = field.widget_options

        if "allowed_values" in options:
            if "remap" in options:
                remap = options["remap"]
                cur_text = remap[config_val]
            else:
                cur_text = config_val
            combo.setCurrentIndex(combo.findText(cur_text))
        else:
            combo.setCurrentIndex(-1)

    def field_changed(self, config_key, field, val, orig_val):
        # TODO support arbitrary nesting?
        try:
            val = field.from_string(val)
            if config_key[0] not in self._changed_config:
                self._changed_config[config_key[0]] = {}
            self._changed_config[config_key[0]][config_key[1]] = val
            self._changed_fields[config_key] = field
            self.fieldChanged.emit(self)
        except:
            widget = self.field_widgets[config_key]
            self.set_value(widget, field, orig_val)




#!/usr/bin/env python

# Demonstrate how to use the vtkBoxWidget to translate, scale, and
# rotate actors.  The basic idea is that the box widget controls an
# actor's transform. A callback which modifies the transform is
# invoked as the box widget is manipulated.

import vtk

# Start by creating some simple geometry; in this case a mace.
sphere = vtk.vtkSphereSource()
cone = vtk.vtkConeSource()
glyph = vtk.vtkGlyph3D()
glyph.SetInputConnection(sphere.GetOutputPort())
glyph.SetSource(cone.GetOutput())
glyph.SetVectorModeToUseNormal()
glyph.SetScaleModeToScaleByVector()
glyph.SetScaleFactor(0.25)
appendData = vtk.vtkAppendPolyData()
appendData.AddInput(glyph.GetOutput())
appendData.AddInput(sphere.GetOutput())
maceMapper = vtk.vtkPolyDataMapper()
maceMapper.SetInputConnection(appendData.GetOutputPort())
maceActor = vtk.vtkLODActor()
maceActor.SetMapper(maceMapper)
maceActor.VisibilityOn()

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# The box widget observes the events invoked by the render window
# interactor.  These events come from user interaction in the render
# window.
boxWidget = vtk.vtkBoxWidget()
boxWidget.SetInteractor(iren)
boxWidget.SetPlaceFactor(1.25)

# Add the actors to the renderer, set the background and window size.
ren.AddActor(maceActor)
ren.SetBackground(0.1, 0.2, 0.4)
renWin.SetSize(300, 300)

# As the box widget is interacted with, it produces a transformation
# matrix that is set on the actor.
t = vtk.vtkTransform()
def TransformActor(obj, event):
    global t, maceActor
    obj.GetTransform(t)
    maceActor.SetUserTransform(t)

# Place the interactor initially. The actor is used to place and scale
# the interactor. An observer is added to the box widget to watch for
# interaction events. This event is captured and used to set the
# transformation matrix of the actor.
boxWidget.SetProp3D(maceActor)
boxWidget.PlaceWidget()
boxWidget.AddObserver("InteractionEvent", TransformActor)

iren.Initialize() 
renWin.Render()
iren.Start()
